"""
Tests the server compute capabilities.
"""

import pytest

import qcfractal.interface as ptl
from qcfractal import testing
from qcfractal.testing import fractal_compute_server, test_server, reset_server_database, using_psi4, using_rdkit

bad_id1 = "000000000000000000000000"
bad_id2 = "000000000000000000000001"


def get_manager_name(fractal_compute_server):

    manager = fractal_compute_server.storage.get_managers()["data"]
    if len(manager) == 0:
        fractal_compute_server.storage.manager_update("test manager")
        return "test manager"
    else:
        return manager[0]["name"]


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(("psi4", "HF", "sto-3g"), id="psi4", marks=using_psi4),
        pytest.param(("rdkit", "UFF", None), id="rdkit", marks=using_rdkit),
    ],
)
def test_task_molecule_no_orientation(data, fractal_compute_server):
    """
    Molecule orientation should not change on compute
    """

    # Reset database each run
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    mol = ptl.Molecule(symbols=["H", "H"], geometry=[0, 0, 0, 0, 5, 0], connectivity=[(0, 1, 1)])

    mol_id = client.add_molecules([mol])[0]

    program, method, basis = data
    ret = client.add_compute(program, method, basis, "energy", None, [mol_id])
    assert "nsubmitted" in str(ret)

    # Manually handle the compute
    fractal_compute_server.await_results()

    # Check for the single result
    ret = client.query_results(id=ret.submitted)
    assert len(ret) == 1
    assert ret[0].status == "COMPLETE"
    assert ret[0].molecule == mol_id

    # Make sure no other molecule was added
    mols = client.query_molecules(molecular_formula=["H2"])
    assert len(mols) == 1
    assert mols[0].id == mol_id

    # Check get_molecule
    mol = ret[0].get_molecule()
    assert ret[0].id == mol_id


@testing.using_rdkit
def test_task_error(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    mol = ptl.models.Molecule(**{"geometry": [0, 0, 0], "symbols": ["He"]})
    # Cookiemonster is an invalid method
    ret = client.add_compute("rdkit", "cookiemonster", "", "energy", None, [mol])

    # Manually handle the compute
    fractal_compute_server.await_results()

    # Check for error
    results = client.query_results(id=ret.submitted)
    assert len(results) == 1
    assert results[0].status == "ERROR"

    assert "connectivity" in results[0].get_error().error_message

    # Check manager
    m = fractal_compute_server.storage.get_managers()["data"]
    assert len(m) == 1
    assert m[0]["failures"] > 0
    assert m[0]["completed"] > 0


@testing.using_rdkit
def test_task_client_restart(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    mol = ptl.models.Molecule(**{"geometry": [0, 0, 1], "symbols": ["He"]})
    # Cookiemonster is an invalid method
    ret = client.add_compute("rdkit", "cookiemonster", "", "energy", None, [mol])

    # Manually handle the compute
    fractal_compute_server.await_results()

    tasks = client.query_tasks(base_result=ret.submitted)[0]
    assert tasks.status == "ERROR"

    upd = client.modify_tasks("restart", ret.submitted)
    assert upd.n_updated == 1

    tasks = client.query_tasks(base_result=ret.submitted)[0]
    assert tasks.status == "WAITING"


@testing.using_rdkit
@testing.using_geometric
def test_task_regenerate(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    # Add a single computation and a geometry optimization
    # Both of these have invalid methods
    mol = ptl.models.Molecule(**{"geometry": [1, 2, 3], "symbols": ["Ne"]})
    geometric_options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "cookiemonster", "basis": "", "keywords": None, "program": "rdkit"},
    }

    ret1 = client.add_compute("rdkit", "cookiemonster", "", "energy", None, [mol])
    ret2 = client.add_procedure("optimization", "geometric", geometric_options, [mol])
    fractal_compute_server.await_results()

    base_ids = [ret1.submitted[0], ret2.submitted[0]]
    old_tasks = client.query_tasks(base_result=base_ids)

    # Regenerate, but old one exists. Should be a no-op
    upd = client.modify_tasks("regenerate", base_result=base_ids)
    assert upd.n_updated == 0
    new_tasks = client.query_tasks(base_result=base_ids)

    for old_task, new_task in zip(old_tasks, new_tasks):
        assert old_task.status == "ERROR"
        assert old_task.id == new_task.id
        assert old_task.base_result == new_task.base_result
        assert old_task.modified_on == new_task.modified_on
        assert old_task.created_on == new_task.created_on

    # Manually delete the old task
    db = fractal_compute_server.objects["storage_socket"]
    db.del_tasks([x.id for x in old_tasks])

    # Actually deleted?
    del_task = client.query_tasks(base_result=base_ids)
    assert len(del_task) == 0

    # Now regenerate
    upd = client.modify_tasks("regenerate", base_result=base_ids)
    new_tasks = client.query_tasks(base_result=base_ids)
    assert upd.n_updated == 2
    for old_task, new_task in zip(old_tasks, new_tasks):
        assert new_task.status == "WAITING"
        assert old_task.id != new_task.id  # Task ids must now be different
        assert old_task.base_result == new_task.base_result
        assert old_task.modified_on < new_task.modified_on  # New task must be newer
        assert old_task.created_on < new_task.created_on  # New task must be newer

    assert old_tasks[0].spec.args[0]["molecule"]["id"] == new_tasks[0].spec.args[0]["molecule"]["id"]
    assert (
        old_tasks[0].spec.args[0]["molecule"]["identifiers"]["molecule_hash"]
        == new_tasks[0].spec.args[0]["molecule"]["identifiers"]["molecule_hash"]
    )
    assert old_tasks[1].spec.args[0]["initial_molecule"]["id"] == new_tasks[1].spec.args[0]["initial_molecule"]["id"]
    assert (
        old_tasks[1].spec.args[0]["initial_molecule"]["identifiers"]["molecule_hash"]
        == new_tasks[1].spec.args[0]["initial_molecule"]["identifiers"]["molecule_hash"]
    )

    # The status of the result should be reset to incomplete
    res = client.query_procedures(base_ids)
    assert all(x.status == "INCOMPLETE" for x in res)


def test_task_modify(test_server):
    client = ptl.FractalClient(test_server)

    # Add a single computation
    mol = ptl.models.Molecule(**{"geometry": [0, 0, 1], "symbols": ["He"]})
    ret1 = client.add_compute("rdkit", "cookiemonster", "", "energy", None, [mol], tag="test_tag_1", priority=1)
    base_id = ret1.submitted[0]
    old_task = client.query_tasks(base_result=base_id)[0]

    assert old_task.priority == 1
    assert old_task.tag == "test_tag_1"

    # Modify the priority and tag
    upd = client.modify_tasks("modify", base_result=base_id, new_tag="test_tag_2", new_priority=0)
    assert upd.n_updated == 1
    new_task = client.query_tasks(base_result=base_id)[0]

    assert new_task.tag == "test_tag_2"
    assert new_task.priority == 0


@testing.using_rdkit
def test_queue_error(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    hooh = ptl.data.get_molecule("hooh.json").copy(update={"connectivity_": None})
    compute_ret = client.add_compute("rdkit", "UFF", "", "energy", None, hooh)

    # Pull out a special iteration on the queue manager
    fractal_compute_server.update_tasks()
    assert len(fractal_compute_server.list_current_tasks()) == 1

    fractal_compute_server.await_results()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    # Pull from database, raw JSON
    db = fractal_compute_server.objects["storage_socket"]
    queue_ret = db.get_queue(status="ERROR")["data"]
    result = db.get_results(id=compute_ret.ids)["data"][0]

    assert len(queue_ret) == 1
    # TODO: task.error is not used anymore
    # assert "connectivity graph" in queue_ret[0].error.error_message
    assert result["status"] == "ERROR"

    # Force a complete mark and test
    fractal_compute_server.objects["storage_socket"].queue_mark_complete([queue_ret[0].id])
    result = db.get_results(id=compute_ret.ids)["data"][0]
    assert result["status"] == "COMPLETE"


@testing.using_rdkit
def test_queue_duplicate_compute(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    hooh = ptl.data.get_molecule("hooh.json")
    mol_ret = client.add_molecules([hooh])

    ret = client.add_compute("rdkit", "UFF", "", "energy", None, mol_ret)
    assert len(ret.ids) == 1
    assert len(ret.existing) == 0

    # Wait for the compute to execute
    fractal_compute_server.await_results()

    db = fractal_compute_server.objects["storage_socket"]

    # Should catch duplicates both ways
    ret = client.add_compute("RDKIT", "uff", None, "energy", None, mol_ret)
    assert len(ret.ids) == 1
    assert len(ret.existing) == 1

    ret = client.add_compute("rdkit", "uFf", "", "energy", None, mol_ret)
    assert len(ret.ids) == 1
    assert len(ret.existing) == 1

    # Multiple queries
    assert len(client.query_results(program="RDKIT")) == 1
    assert len(client.query_results(program="RDKit")) == 1

    assert len(client.query_results(method="UFF")) == 1
    assert len(client.query_results(method="uff")) == 1

    assert len(client.query_results(basis=None)) == 1
    assert len(client.query_results(basis="")) == 1

    assert len(client.query_results(keywords=None)) == 1


@testing.using_rdkit
def test_queue_compute_mixed_molecule(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.1")
    mol_ret = client.add_molecules([mol1])

    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")

    ret = client.add_compute("RDKIT", "UFF", "", "energy", None, [mol1, mol2, bad_id1], full_return=True)
    assert len(ret.data.ids) == 3
    assert ret.data.ids[2] is None
    assert len(ret.data.submitted) == 2
    assert len(ret.data.existing) == 0

    # Pull out fireworks launchpad and queue nanny
    fractal_compute_server.await_results()

    db = fractal_compute_server.objects["storage_socket"]

    ret = client.add_compute("rdkit", "UFF", "", "energy", None, [mol_ret[0], bad_id2])
    assert len(ret.ids) == 2
    assert ret.ids[1] is None
    assert len(ret.submitted) == 0
    assert len(ret.existing) == 1


@testing.using_rdkit
@testing.using_geometric
def test_queue_duplicate_procedure(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    hooh = ptl.data.get_molecule("hooh.json")
    mol_ret = client.add_molecules([hooh])

    geometric_options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
    }

    ret = client.add_procedure("optimization", "geometric", geometric_options, [mol_ret[0], bad_id1])
    assert len(ret.ids) == 2
    assert ret.ids[1] is None
    assert len(ret.submitted) == 1
    assert len(ret.existing) == 0

    # Pull out fireworks launchpad and queue nanny
    fractal_compute_server.await_results()

    db = fractal_compute_server.objects["storage_socket"]

    ret2 = client.add_procedure("optimization", "geometric", geometric_options, [bad_id1, hooh])
    assert len(ret2.ids) == 2
    assert ret2.ids[0] is None
    assert len(ret2.submitted) == 0
    assert len(ret2.existing) == 1

    assert ret.ids[0] == ret2.ids[1]


def test_queue_bad_compute_method(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.1")

    with pytest.raises(IOError) as exc:
        ret = client.add_compute("badprogram", "UFF", "", "energy", None, [mol1], full_return=True)

    assert "not avail" in str(exc.value)


def test_queue_bad_procedure_method(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)
    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.1")

    geometric_options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
    }

    # Test bad procedure
    with pytest.raises(IOError) as exc:
        ret = client.add_procedure("optimization", "badproc", geometric_options, [mol1])

    assert "not avail" in str(exc.value)

    # Test procedure class
    with pytest.raises(IOError) as exc:
        ret = client.add_procedure("badprocedure", "geometric", geometric_options, [mol1])

    assert "Unknown procedure" in str(exc.value)

    # Test bad program
    with pytest.raises(IOError) as exc:
        geometric_options["qc_spec"]["program"] = "badqc"
        ret = client.add_procedure("optimization", "geometric", geometric_options, [mol1])

    assert "not avail" in str(exc.value)
    assert "badqc" in str(exc.value)


def test_queue_ordering_time(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 1.1")
    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")

    ret1 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol1).ids[0]
    ret2 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol2).ids[0]

    manager = get_manager_name(fractal_compute_server)

    assert len(fractal_compute_server.storage.queue_get_next(manager, [], [], limit=1)) == 0

    queue_id1 = fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], [], limit=1)[0].base_result
    queue_id2 = fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], [], limit=1)[0].base_result

    assert queue_id1 == ret1
    assert queue_id2 == ret2


def test_queue_ordering_priority(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 1.1")
    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")
    mol3 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 3.3")

    ret1 = client.add_compute("rdkit", "uff", "", "energy", None, mol1).ids[0]
    ret2 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol2, priority="high").ids[0]
    ret3 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol3, priority="HIGH").ids[0]

    manager = get_manager_name(fractal_compute_server)

    queue_id1 = fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], [], limit=1)[0].base_result
    queue_id2 = fractal_compute_server.storage.queue_get_next(manager, ["RDkit"], [], limit=1)[0].base_result
    queue_id3 = fractal_compute_server.storage.queue_get_next(manager, ["RDKIT"], [], limit=1)[0].base_result

    assert queue_id1 == ret2
    assert queue_id2 == ret3
    assert queue_id3 == ret1


def test_queue_order_procedure_priority(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    geometric_options = {
        "keywords": None,
        "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
    }

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 1.1")
    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")
    mol3 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 3.3")

    ret1 = client.add_procedure("optimization", "geometric", geometric_options, [mol1]).ids[0]
    ret2 = client.add_procedure("OPTIMIZATION", "geometric", geometric_options, [mol2], priority="high").ids[0]
    ret3 = client.add_procedure("OPTimization", "GEOmetric", geometric_options, [mol3], priority="HIGH").ids[0]

    manager = get_manager_name(fractal_compute_server)

    assert len(fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], [], limit=1)) == 0
    assert len(fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], ["geom"], limit=1)) == 0
    assert len(fractal_compute_server.storage.queue_get_next(manager, ["prog1"], ["geometric"], limit=1)) == 0

    queue_id1 = fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], ["geometric"], limit=1)[0].base_result
    queue_id2 = fractal_compute_server.storage.queue_get_next(manager, ["RDKIT"], ["geometric"], limit=1)[0].base_result
    queue_id3 = fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], ["GEOMETRIC"], limit=1)[0].base_result

    assert queue_id1 == ret2
    assert queue_id2 == ret3
    assert queue_id3 == ret1


def test_queue_query_tag(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 1.1")
    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")
    mol3 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 3.3")

    ret1 = client.add_compute("rdkit", "uff", "", "energy", None, mol1).ids[0]
    ret2 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol2, tag="test").ids[0]
    ret3 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol3, tag="test2").ids[0]

    tasks_tag_test = client.query_tasks(tag="test")
    assert len(tasks_tag_test) == 1
    assert tasks_tag_test[0].base_result == ret2

    tasks_tag_none = client.query_tasks()
    assert len(tasks_tag_none) == 3
    assert {task.base_result for task in tasks_tag_none} == {ret1, ret2, ret3}

    tasks_tagged = client.query_tasks(tag=["test", "test2"])
    assert len(tasks_tagged) == 2
    assert {task.base_result for task in tasks_tagged} == {ret2, ret3}


def test_queue_query_manager(fractal_compute_server):
    reset_server_database(fractal_compute_server)

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 1.1")
    mol2 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 2.2")
    mol3 = ptl.Molecule.from_data("He 0 0 0\nHe 0 0 3.3")

    ret1 = client.add_compute("rdkit", "uff", "", "energy", None, mol1).ids[0]
    ret2 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol2).ids[0]
    ret3 = client.add_compute("RDKIT", "UFF", "", "energy", None, mol3).ids[0]

    manager = get_manager_name(fractal_compute_server)
    fractal_compute_server.storage.queue_get_next(manager, ["rdkit"], [], limit=1)[0]
    tasks_manager = client.query_tasks(manager=manager)
    assert len(tasks_manager) == 1
    assert tasks_manager[0].base_result == ret1

    fractal_compute_server.storage.queue_get_next(manager, ["RDkit"], [], limit=1)[0]
    fractal_compute_server.storage.queue_get_next(manager, ["RDKIT"], [], limit=1)[0]
    tasks_manager = client.query_tasks(manager=manager)
    assert len(tasks_manager) == 3



"""Msgpack Base Results Phase 1

Revision ID: 8b0cd9accaf2
Revises: 1134312ad4a3
Create Date: 2019-08-11 22:30:27.613722

"""
from alembic import op
import sqlalchemy as sa

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers import msgpack_migrations
from qcelemental.util import msgpackext_dumps, msgpackext_loads

# revision identifiers, used by Alembic.
revision = "8b0cd9accaf2"
down_revision = "1134312ad4a3"
branch_labels = None
depends_on = None

block_size = 100
table_name = "base_result"


def transformer(old_data):

    extras = old_data["extras"]
    extras.pop("_qcfractal_tags", None)  # cleanup old tags

    return {"extras_": msgpackext_dumps(extras)}


update_columns = {"extras"}


def upgrade():
    msgpack_migrations.json_to_msgpack_table(table_name, block_size, update_columns, transformer)


def downgrade():
    msgpack_migrations.json_to_msgpack_table_dropcols(table_name, block_size, update_columns)



"""Updates collections table with owner and visibility

Revision ID: a2f76bb7be65
Revises: c194a8ef6acf
Create Date: 2019-10-08 11:06:15.438742

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a2f76bb7be65"
down_revision = "c194a8ef6acf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("collection", sa.Column("description", sa.String(), nullable=True))

    op.add_column("collection", sa.Column("group", sa.String(), nullable=False, server_default="default"))
    op.alter_column("collection", "group", server_default=None)

    op.add_column("collection", sa.Column("view_url_hdf5", sa.String(), nullable=True))
    op.add_column("collection", sa.Column("view_url_plaintext", sa.String(), nullable=True))
    op.add_column("collection", sa.Column("view_metadata", sa.JSON(), nullable=True))

    op.add_column("collection", sa.Column("view_available", sa.Boolean(), nullable=True))
    op.execute("UPDATE collection SET view_available=false")
    op.alter_column("collection", "view_available", nullable=False)

    op.add_column("collection", sa.Column("visibility", sa.Boolean(), nullable=True))
    op.execute("UPDATE collection SET visibility=true")
    op.alter_column("collection", "visibility", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("collection", "visibility")
    op.drop_column("collection", "view_url_hdf5")
    op.drop_column("collection", "view_url_plaintext")
    op.drop_column("collection", "view_metadata")
    op.drop_column("collection", "view_available")
    op.drop_column("collection", "group")
    op.drop_column("collection", "metadata")
    op.drop_column("collection", "description")
    # ### end Alembic commands ###



"""Msgpack Molecule Phase 1

Revision ID: 963822c28879
Revises: 4bb79efa9855
Create Date: 2019-08-10 17:41:15.520300

"""
from alembic import op
import sqlalchemy as sa
import numpy as np

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))

from migration_helpers import msgpack_migrations
from qcelemental.util import msgpackext_dumps, msgpackext_loads

# revision identifiers, used by Alembic.
revision = "963822c28879"
down_revision = "4bb79efa9855"
branch_labels = None
depends_on = None

block_size = 100
table_name = "molecule"

converters = {
    "symbols": lambda arr: np.array(arr, dtype=str),
    "geometry": lambda arr: np.array(arr, dtype=float),
    "masses": lambda arr: np.array(arr, dtype=float),
    "real": lambda arr: np.array(arr, dtype=bool),
    "atom_labels": lambda arr: np.array(arr, dtype=str),
    "atomic_numbers": lambda arr: np.array(arr, dtype=np.int16),
    "mass_numbers": lambda arr: np.array(arr, dtype=np.int16),
    "fragments": lambda list_arr: [np.array(x, dtype=np.int32) for x in list_arr],
}


def transformer(old_data):

    row = {}
    for k, v in old_data.items():
        if k == "id":
            continue
        d = msgpackext_dumps(converters[k](v))
        row[k + "_"] = d

    return row


def upgrade():
    msgpack_migrations.json_to_msgpack_table(table_name, block_size, converters.keys(), transformer)


def downgrade():
    msgpack_migrations.json_to_msgpack_table_dropcols(table_name, block_size, update_columns)



"""
Python init for QCPortal collections
"""

from .collection_utils import collection_factory, collections_name_map, list_known_collections, register_collection
from .dataset import Dataset
from .dataset_view import DatasetView, HDF5View, PlainTextView, RemoteView
from .generic import Generic
from .gridoptimization_dataset import GridOptimizationDataset
from .optimization_dataset import OptimizationDataset
from .reaction_dataset import ReactionDataset
from .torsiondrive_dataset import TorsionDriveDataset



"""
Tests the QCPortal dataset object
"""

import pytest

from . import portal
from . import test_helper as th


def _compare_stoichs(stoich, stoich_other):
    mols = list(stoich)
    mols_other = list(stoich_other)
    assert th.compare_lists(mols, mols_other)

    for mol in mols:
        assert stoich[mol] == stoich_other[mol]

    return True


def _compare_rxn_stoichs(ref, new):
    stoich = ref.stoichiometry
    stoich_other = new.stoichiometry

    keys = list(stoich)
    keys_other = list(stoich_other)
    assert th.compare_lists(keys, keys_other)

    for k in keys:
        _compare_stoichs(stoich[k], stoich_other[k])

    return True


# Build an interesting dataset
@pytest.fixture
def water_ds():
    # Create water ReactionDataset, also tests that ds_type is case insensitive
    ds = portal.collections.ReactionDataset("Water Data", ds_type="RxN")

    # Build the water dimer.
    dimer = portal.data.get_molecule("water_dimer_minima.psimol")
    frag_0 = dimer.get_fragment(0, orient=True, group_fragments=True)
    frag_1 = dimer.get_fragment(1, orient=True, group_fragments=True)
    frag_0_1 = dimer.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = dimer.get_fragment(1, 0, orient=True, group_fragments=True)

    # Add single stoich rxn via list
    ds.add_rxn(
        "Water Dimer, nocp",
        [(dimer, 1.0), (frag_0, -1.0), (frag_1, -1.0)],
        attributes={"R": "Minima"},
        reaction_results={"Benchmark": -20.0, "DFT": -10.0},
    )

    dimer_string = dimer.to_string("psi4")
    # Add single stoich from strings, not a valid set
    ds.add_rxn(
        "Water Dimer, dimer - str (invalid)",
        [(dimer_string, 1.0), (frag_0, 0.0)],
        attributes={"R": "Minima"},
        reaction_results={"Benchmark": -20.0, "DFT": -10.0},
    )

    # Add single stoich rxn via hashes
    ds.add_rxn(
        "Water Dimer, nocp - hash",
        [(dimer.get_hash(), 1.0), (frag_0.get_hash(), -1.0), (frag_1.get_hash(), -1.0)],
        attributes={"R": "Minima"},
        reaction_results={"Benchmark": -5.0},
    )

    # Add multi stoich reaction via dict
    with pytest.raises(KeyError):
        ds.add_rxn("Null", {"Null": [(dimer, 1.0)]})

    # nocp and cp water dimer
    ds.add_rxn(
        "Water Dimer, all",
        {
            "cp": [(dimer, 1.0), (frag_0_1, -1.0), (frag_1_0, -1.0)],
            "default": [(dimer, 1.0), (frag_0, -1.0), (frag_1, -1.0)],
        },
        other_fields={"Something": "Other thing"},
    )

    ds.add_ie_rxn("Water dimer", dimer.to_string("psi4"))

    # Add unverified records (requires a active server)
    ds.data.__dict__["records"] = ds._new_records

    return ds


# Build a nbody dataset
@pytest.fixture
def nbody_ds():
    ds = portal.collections.ReactionDataset("N-Body Data")

    dimer = portal.data.get_molecule("water_dimer_minima.psimol")
    frag_0 = dimer.get_fragment(0, orient=True, group_fragments=True)
    frag_1 = dimer.get_fragment(1, orient=True, group_fragments=True)
    frag_0_1 = dimer.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = dimer.get_fragment(1, 0, orient=True, group_fragments=True)

    ds.add_rxn(
        "Water Dimer, bench",
        {
            "cp1": [(frag_0_1, 1.0), (frag_1_0, 1.0)],
            "default1": [(frag_0, 1.0), (frag_1, 1.0)],
            "cp": [(dimer, 1.0)],
            "default": [(dimer, 1.0)],
        },
    )

    ds.add_ie_rxn("Water Dimer", dimer.to_string("psi4"))
    ds.add_ie_rxn("Ne Tetramer", portal.data.get_molecule("neon_tetramer.psimol"))

    # Ne Tetramer benchmark
    ds.ne_stoich = {
        "name": "Ne Tetramer",
        "stoichiometry": {
            "default1": {"acfce0c62cac57640d221a545998711853606bc9": 4.0},
            "cp1": {
                "532ae1b0fab346aeaa2167972b5b61079081209f": 1.0,
                "f89a033c52c527b0a85f271193f3515b0aec190b": 1.0,
                "43fafa29d4d3197738e741051aff71ffe98264b3": 1.0,
                "cf36366bf4d5a2f96cf0669425c49bd929d7f081": 1.0,
            },
            "default2": {
                "acfce0c62cac57640d221a545998711853606bc9": -8.0,
                "e643e8f4e8668793165339e98f4f314cdac6e909": 1.0,
                "0768204c7ddc575a77574b49598b89eb04d4522b": 1.0,
                "05cb2410011128a45b5c3479e7724c60b91bfbee": 2.0,
                "444dba64349a093f8da2c5c6594b128f54619d92": 1.0,
                "75608da1aad0be8f353cab54e2674a138c5d04e7": 1.0,
            },
            "cp2": {
                "532ae1b0fab346aeaa2167972b5b61079081209f": -2.0,
                "f89a033c52c527b0a85f271193f3515b0aec190b": -2.0,
                "43fafa29d4d3197738e741051aff71ffe98264b3": -2.0,
                "cf36366bf4d5a2f96cf0669425c49bd929d7f081": -2.0,
                "f08a3ea572f4a371c18b99be63545a23f9cf17d2": 1.0,
                "eb6efb5f6a7238c0582517eeba0c5b940d84c51d": 1.0,
                "75a23e5e6f995aaee9a049226701f543f0ee9a82": 1.0,
                "8d5280d9aa82243a12b319b24bd19506bae77853": 1.0,
                "b7f73f0957f1f14469dde2000605061fbbe4939e": 1.0,
                "bee9dfe443c016283ef2a4d98959a736b0d0b591": 1.0,
            },
            "default3": {
                "acfce0c62cac57640d221a545998711853606bc9": 4.0,
                "e643e8f4e8668793165339e98f4f314cdac6e909": -1.0,
                "0768204c7ddc575a77574b49598b89eb04d4522b": -1.0,
                "05cb2410011128a45b5c3479e7724c60b91bfbee": -2.0,
                "444dba64349a093f8da2c5c6594b128f54619d92": -1.0,
                "75608da1aad0be8f353cab54e2674a138c5d04e7": -1.0,
                "da4d287f648d33203652cd5a8fc621e331c16fab": 1.0,
                "d7914b7213eff227af8228f64c26187eec343c21": 1.0,
                "efcfd9e9ae9f33dcfb2780367690dcfff8bcb397": 1.0,
                "4e3e461ab1439dc360a4b001ac885a8240b27640": 1.0,
            },
            "cp3": {
                "532ae1b0fab346aeaa2167972b5b61079081209f": 1.0,
                "f89a033c52c527b0a85f271193f3515b0aec190b": 1.0,
                "43fafa29d4d3197738e741051aff71ffe98264b3": 1.0,
                "cf36366bf4d5a2f96cf0669425c49bd929d7f081": 1.0,
                "f08a3ea572f4a371c18b99be63545a23f9cf17d2": -1.0,
                "eb6efb5f6a7238c0582517eeba0c5b940d84c51d": -1.0,
                "75a23e5e6f995aaee9a049226701f543f0ee9a82": -1.0,
                "8d5280d9aa82243a12b319b24bd19506bae77853": -1.0,
                "b7f73f0957f1f14469dde2000605061fbbe4939e": -1.0,
                "bee9dfe443c016283ef2a4d98959a736b0d0b591": -1.0,
                "32b290f24926d5ce6ca7ed4eb31d034bcc18a784": 1.0,
                "214a0c4469f1ea514e0ee7a256dc601f39a90498": 1.0,
                "fb3eeb93515fac8bb807af999bddd883c27af7fa": 1.0,
                "3e4f560822d53d4fd53520d354eb4fa6632c745a": 1.0,
            },
            "default": {"c0e6a8225f1eb3ab5f7e81a755a91dd800869f23": 1.0},
            "cp": {"c0e6a8225f1eb3ab5f7e81a755a91dd800869f23": 1.0},
        },
        "attributes": {},
        "reaction_results": {"default": {}},
    }

    # Add unverified records (requires a active server)
    ds.data.__dict__["records"] = ds._new_records

    return ds


# Test conventional add
def test_rxn_add(water_ds):

    assert water_ds.data.name == "Water Data"
    assert len(water_ds.get_index()) == 5

    nocp_stoich_class = water_ds.get_rxn("Water Dimer, nocp").stoichiometry["default"]
    nocp_stoich_hash = water_ds.get_rxn("Water Dimer, nocp - hash").stoichiometry["default"]
    nocp_stoich_dict = water_ds.get_rxn("Water Dimer, all").stoichiometry["default"]

    # Check if both builds check out
    _compare_stoichs(nocp_stoich_class, nocp_stoich_hash)
    _compare_stoichs(nocp_stoich_class, nocp_stoich_dict)


# Test IE add
def test_nbody_rxn(nbody_ds):

    # Check the Water Dimer
    water_stoich_bench = nbody_ds.get_rxn("Water Dimer, bench")
    water_stoich = nbody_ds.get_rxn("Water Dimer")
    _compare_rxn_stoichs(water_stoich, water_stoich_bench)

    bench_vals = {"default1": 1, "cp1": 4, "default2": 6, "cp2": 10, "default3": 10, "cp3": 14, "default": 1, "cp": 1}
    # Check some basics
    for key in list(nbody_ds.ne_stoich["stoichiometry"]):
        assert bench_vals[key] == len(nbody_ds.ne_stoich["stoichiometry"][key])

    # Check the N-body
    ne_stoich = nbody_ds.get_rxn("Ne Tetramer")
    mh = list(ne_stoich.stoichiometry["default"])[0]
    # print(ne_stoich)
    # _compare_rxn_stoichs(nbody_ds.ne_stoich, ne_stoich)


def test_database_history():
    ds = portal.collections.Dataset("history_test")
    history = [
        ("energy", "p3", "m1", "b2", "o1"),
        ("energy", "p1", "m1", None, "o1"),
        ("energy", "p1", "m1", None, "o2"),
        ("energy", "p1", "m2", "b3", "o1"),
        ("gradient", "p1", "m2", None, None),
    ]  # yapf: disable

    for h in history:
        ds._add_history(driver=h[0], program=h[1], method=h[2], basis=h[3], keywords=h[4])

    assert ds.list_records().shape[0] == 5
    assert ds.list_records(program="P1").shape[0] == 4
    assert ds.list_records(basis="None").shape[0] == 3
    assert ds.list_records(keywords="None").shape[0] == 1



"""
Common models for QCPortal/Fractal
"""
import json

# For compression
import lzma
import bz2
import gzip

from enum import Enum
from typing import Any, Dict, Optional, Union

from pydantic import Field, validator
from qcelemental.models import AutodocBaseSettings, Molecule, ProtoModel, Provenance
from qcelemental.models.procedures import OptimizationProtocols
from qcelemental.models.results import ResultProtocols

from .model_utils import hash_dictionary, prepare_basis, recursive_normalizer

__all__ = ["QCSpecification", "OptimizationSpecification", "KeywordSet", "ObjectId", "DriverEnum", "Citation"]

# Add in QCElemental models
__all__.extend(["Molecule", "Provenance", "ProtoModel"])

# Autodoc
__all__.extend(["OptimizationProtocols", "ResultProtocols"])


class ObjectId(str):
    """
    The Id of the object in the data.
    """

    _valid_hex = set("0123456789abcdef")

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, str) and (len(v) == 24) and (set(v) <= cls._valid_hex):
            return v
        elif isinstance(v, int):
            return str(v)
        elif isinstance(v, str) and v.isdigit():
            return v
        else:
            raise TypeError("The string {} is not a valid 24-character hexadecimal or integer ObjectId!".format(v))


class DriverEnum(str, Enum):
    """
    The type of calculation that is being performed (e.g., energy, gradient, Hessian, ...).
    """

    energy = "energy"
    gradient = "gradient"
    hessian = "hessian"
    properties = "properties"


class CompressionEnum(str, Enum):
    """
    How data is compressed (compression method only, ie gzip, bzip2)
    """

    none = "none"
    gzip = "gzip"
    bzip2 = "bzip2"
    lzma = "lzma"


class KVStore(ProtoModel):
    """
    Storage of outputs and error messages, with optional compression
    """

    id: int = Field(
        None, description="Id of the object on the database. This is assigned automatically by the database."
    )

    compression: CompressionEnum = Field(CompressionEnum.none, description="Compression method (such as gzip)")
    compression_level: int = Field(0, description="Level of compression (typically 0-9)")
    data: bytes = Field(..., description="Compressed raw data of output/errors, etc")

    @validator("data", pre=True)
    def _set_data(cls, data, values):
        """Handles special data types

        Strings are converted to byte arrays, and dicts are converted via json.dumps. If a string or
        dictionary is given, then compression & compression level must be none/0 (the defaults)

        Will chack that compression and compression level are None/0. Since this validator
        runs after all the others, that is safe.

        (According to pydantic docs, validators are run in the order of field definition)
        """
        if isinstance(data, dict):
            if values["compression"] != CompressionEnum.none:
                raise ValueError("Compression is set, but input is a dictionary")
            if values["compression_level"] != 0:
                raise ValueError("Compression level is set, but input is a dictionary")
            return json.dumps(data).encode()
        elif isinstance(data, str):
            if values["compression"] != CompressionEnum.none:
                raise ValueError("Compression is set, but input is a string")
            if values["compression_level"] != 0:
                raise ValueError("Compression level is set, but input is a string")
            return data.encode()
        else:
            return data

    @validator("compression", pre=True)
    def _set_compression(cls, compression):
        """Sets the compression type to CompressionEnum.none if compression is None

        Needed as older entries in the database have null for compression/compression_level
        """
        if compression is None:
            return CompressionEnum.none
        else:
            return compression

    @validator("compression_level", pre=True)
    def _set_compression_level(cls, compression_level):
        """Sets the compression_level to zero if compression is None

        Needed as older entries in the database have null for compression/compression_level
        """
        if compression_level is None:
            return 0
        else:
            return compression_level

    @classmethod
    def compress(
        cls,
        input_data: Union[Dict[str, str], str],
        compression_type: CompressionEnum = CompressionEnum.none,
        compression_level: Optional[int] = None,
    ):
        """Compresses a string given a compression scheme and level

        Returns an object of type `cls`

        If compression_level is None, but a compression_type is specified, an appropriate default level is chosen
        """

        if isinstance(input_data, dict):
            input_data = json.dumps(input_data)

        data = input_data.encode()

        # No compression
        if compression_type is CompressionEnum.none:
            compression_level = 0

        # gzip compression
        elif compression_type is CompressionEnum.gzip:
            if compression_level is None:
                compression_level = 6
            data = gzip.compress(data, compresslevel=compression_level)

        # bzip2 compression
        elif compression_type is CompressionEnum.bzip2:
            if compression_level is None:
                compression_level = 6
            data = bz2.compress(data, compresslevel=compression_level)

        # LZMA compression
        # By default, use level = 1 for larger files (>15MB or so)
        elif compression_type is CompressionEnum.lzma:
            if compression_level is None:
                if len(data) > 15 * 1048576:
                    compression_level = 1
                else:
                    compression_level = 6
            data = lzma.compress(data, preset=compression_level)
        else:
            # Shouldn't ever happen, unless we change CompressionEnum but not the rest of this function
            raise TypeError("Unknown compression type??")

        return cls(data=data, compression=compression_type, compression_level=compression_level)

    def get_string(self):
        """
        Returns the string representing the output
        """
        if self.compression is CompressionEnum.none:
            return self.data.decode()
        elif self.compression is CompressionEnum.gzip:
            return gzip.decompress(self.data).decode()
        elif self.compression is CompressionEnum.bzip2:
            return bz2.decompress(self.data).decode()
        elif self.compression is CompressionEnum.lzma:
            return lzma.decompress(self.data).decode()
        else:
            # Shouldn't ever happen, unless we change CompressionEnum but not the rest of this function
            raise TypeError("Unknown compression type??")

    def get_json(self):
        """
        Returns a dict if the data stored is a JSON string

        (errors are stored as JSON. stdout/stderr are just strings)
        """
        s = self.get_string()
        return json.loads(s)


class QCSpecification(ProtoModel):
    """
    The quantum chemistry metadata specification for individual computations such as energy, gradient, and Hessians.
    """

    driver: DriverEnum = Field(..., description=str(DriverEnum.__doc__))
    method: str = Field(..., description="The quantum chemistry method to evaluate (e.g., B3LYP, PBE, ...).")
    basis: Optional[str] = Field(
        None,
        description="The quantum chemistry basis set to evaluate (e.g., 6-31g, cc-pVDZ, ...). Can be ``None`` for "
        "methods without basis sets.",
    )
    keywords: Optional[ObjectId] = Field(
        None,
        description="The Id of the :class:`KeywordSet` registered in the database to run this calculation with. This "
        "Id must exist in the database.",
    )
    protocols: Optional[ResultProtocols] = Field(ResultProtocols(), description=str(ResultProtocols.__base_doc__))
    program: str = Field(
        ...,
        description="The quantum chemistry program to evaluate the computation with. Not all quantum chemistry programs"
        " support all combinations of driver/method/basis.",
    )

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)

        # Maintain hash compatability
        if len(ret["protocols"]) == 0:
            ret.pop("protocols", None)

        return ret

    @validator("basis")
    def _check_basis(cls, v):
        return prepare_basis(v)

    @validator("program")
    def _check_program(cls, v):
        return v.lower()

    @validator("method")
    def _check_method(cls, v):
        return v.lower()


class OptimizationSpecification(ProtoModel):
    """
    Metadata describing a geometry optimization.
    """

    program: str = Field(..., description="Optimization program to run the optimization with")
    keywords: Optional[Dict[str, Any]] = Field(
        None,
        description="Dictionary of keyword arguments to pass into the ``program`` when the program runs. "
        "Note that unlike :class:`QCSpecification` this is a dictionary of keywords, not the Id for a "
        ":class:`KeywordSet`. ",
    )
    protocols: Optional[OptimizationProtocols] = Field(
        OptimizationProtocols(), description=str(OptimizationProtocols.__base_doc__)
    )

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)

        # Maintain hash compatability
        if len(ret["protocols"]) == 0:
            ret.pop("protocols", None)

        return ret

    @validator("program")
    def _check_program(cls, v):
        return v.lower()

    @validator("keywords")
    def _check_keywords(cls, v):
        if v is not None:
            v = recursive_normalizer(v)
        return v


class KeywordSet(ProtoModel):
    """
    A key:value storage object for Keywords.
    """

    id: Optional[ObjectId] = Field(
        None, description="The Id of this object, will be automatically assigned when added to the database."
    )
    hash_index: str = Field(
        ...,
        description="The hash of this keyword set to store and check for collisions. This string is automatically "
        "computed.",
    )
    values: Dict[str, Optional[Any]] = Field(
        ...,
        description="The key-value pairs which make up this KeywordSet. There is no direct relation between this "
        "dictionary and applicable program/spec it can be used on.",
    )
    lowercase: bool = Field(
        True,
        description="String keys are in the ``values`` dict are normalized to lowercase if this is True. Assists in "
        "matching against other :class:`KeywordSet` objects in the database.",
    )
    exact_floats: bool = Field(
        False,
        description="All floating point numbers are rounded to 1.e-10 if this is False."
        "Assists in matching against other :class:`KeywordSet` objects in the database.",
    )
    comments: Optional[str] = Field(
        None,
        description="Additional comments for this KeywordSet. Intended for pure human/user consumption " "and clarity.",
    )

    def __init__(self, **data):

        build_index = False
        if ("hash_index" not in data) or data.pop("build_index", False):
            build_index = True
            data["hash_index"] = "placeholder"

        ProtoModel.__init__(self, **data)

        # Overwrite options with massaged values
        kwargs = {"lowercase": self.lowercase}
        if self.exact_floats:
            kwargs["digits"] = False

        self.__dict__["values"] = recursive_normalizer(self.values, **kwargs)

        # Build a hash index if we need it
        if build_index:
            self.__dict__["hash_index"] = self.get_hash_index()

    def get_hash_index(self):
        return hash_dictionary(self.values.copy())


class Citation(ProtoModel):
    """ A literature citation.  """

    acs_citation: Optional[
        str
    ] = None  # hand-formatted citation in ACS style. In the future, this could be bibtex, rendered to different formats.
    bibtex: Optional[str] = None  # bibtex blob for later use with bibtex-renderer
    doi: Optional[str] = None
    url: Optional[str] = None

    def to_acs(self) -> str:
        """ Returns an ACS-formatted citation """
        return self.acs_citation



import abc
import distutils
import hashlib
import pathlib
import shutil
import tarfile
import tempfile
import warnings
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, NoReturn, Optional, Tuple, Union

import numpy as np
import pandas as pd
import h5py
from qcelemental.util.serialization import deserialize, serialize

from ..models import Molecule, ObjectId
from ..util import normalize_filename
from .dataset import Dataset, MoleculeEntry
from .reaction_dataset import ReactionDataset, ReactionEntry

if TYPE_CHECKING:  # pragma: no cover
    from .. import FractalClient
    from ..models.rest_models import CollectionSubresourceGETResponseMeta


class DatasetView(abc.ABC):
    @abc.abstractmethod
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def write(self, ds: Dataset) -> None:
        """
        Writes a dataset to disk.

        Parameters
        ----------
        ds: Dataset
            The dataset to write.

        Returns
        -------
            None
        """

    @abc.abstractmethod
    def list_values(self) -> pd.DataFrame:
        """
        Get a list of all available value columns.

        Returns
        -------
            A Dataframe with specification of available columns.
        """

    @abc.abstractmethod
    def get_values(
        self, queries: List[Dict[str, Union[str, bool]]], subset: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Get value columns.

        Parameters
        ----------
        queries: List[Dict[str, Union[str, bool]]]
            List of column metadata to match.
        subset: Optional[List[str]], optional
            The indices of the desired subset. Return all indices if subset is None.

        Returns
        -------
            A Dataframe whose columns correspond to each query and a dictionary of units for each column.
        """

    @abc.abstractmethod
    def get_molecules(self, indexes: List[Union[ObjectId, int]]) -> pd.Series:
        """
        Get a list of molecules using a molecule indexer

        Parameters
        ----------
        indexes : List['ObjectId']
            A list of molecule ids to return

        Returns
        -------
        pd.Series
            A Series of Molecules corresponding to indexes
        """

    @abc.abstractmethod
    def get_entries(self, subset: Optional[List[str]] = None) -> pd.DataFrame:
        """
        Get a list of entries in the dataset

        Parameters
        ----------
        subset: Optional[List[str]], optional
            The indices of the desired subset. Return all indices if subset is None.

        Returns
        -------
        pd.DataFrame
            A dataframe of entries
        """


class HDF5View(DatasetView):
    def __init__(self, path: Union[str, pathlib.Path]) -> None:
        """
        Parameters
        ----------
        path: Union[str, pathlib.Path]
            File path of view
        """
        path = pathlib.Path(path)
        self._path = path
        self._entries: pd.DataFrame = None
        self._index: pd.DataFrame = None

    def list_values(self) -> pd.DataFrame:
        with self._read_file() as f:
            history_keys = self._deserialize_field(f.attrs["history_keys"])
            df = pd.DataFrame(columns=history_keys + ["name", "native"])
            for dataset in f["value"].values():
                row = {k: self._deserialize_field(dataset.attrs[k]) for k in history_keys}
                row["name"] = self._deserialize_field(dataset.attrs["name"])
                row["native"] = True
                df = df.append(row, ignore_index=True)
            for dataset in f["contributed_value"].values():
                row = dict()
                row["name"] = self._deserialize_field(dataset.attrs["name"])
                for k in history_keys:
                    row[k] = "Unknown"
                # ReactionDataset uses "default" as a default value for stoich, but many contributed datasets lack a stoich field
                if "stoichiometry" in history_keys:
                    row["stoichiometry"] = "default"
                if "theory_level_details" in dataset.attrs:
                    theory_level_details = self._deserialize_field(dataset.attrs["theory_level_details"])
                    if isinstance(theory_level_details, dict):
                        row.update(**theory_level_details)
                row["native"] = False
                df = df.append(row, ignore_index=True)
        # for some reason, pandas makes native a float column
        return df.astype({"native": bool})

    def get_values(
        self, queries: List[Dict[str, Union[str, bool]]], subset: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Parameters
        ----------
        subset
        queries: List[Dict[str, Union[str, bool]]]
            List of queries. Fields actually used are native, name, driver
        """

        units = {}
        entries = self.get_index(subset)
        indexes = entries._h5idx
        with self._read_file() as f:
            ret = pd.DataFrame(index=entries["index"])

            for query in queries:
                dataset_name = "value/" if query["native"] else "contributed_value/"
                dataset_name += self._normalize_hdf5_name(query["name"])
                driver = query["driver"]

                dataset = f[dataset_name]
                if not h5py.check_dtype(vlen=dataset.dtype):
                    data = [dataset[i] for i in indexes]
                else:
                    if driver.lower() == "gradient":
                        data = [np.reshape(dataset[i], (-1, 3)) for i in indexes]
                    elif driver.lower() == "hessian":
                        data = []
                        for i in indexes:
                            n2 = len(dataset[i])
                            n = int(round(np.sqrt(n2)))
                            data.append(np.reshape(dataset[i], (n, n)))
                    else:
                        warnings.warn(
                            f"Variable length data type not understood, returning flat array " f"(driver = {driver}).",
                            RuntimeWarning,
                        )
                        try:
                            data = [np.array(dataset[i]) for i in indexes]
                        except ValueError:
                            data = [dataset[i] for i in indexes]
                column_name = query["name"]
                column_units = self._deserialize_field(dataset.attrs["units"])
                ret[column_name] = data
                units[column_name] = column_units

        return ret, units

    def get_molecules(self, indexes: List[Union[ObjectId, int]], keep_serialized: bool = False) -> pd.Series:
        with self._read_file() as f:
            mol_schema = f["molecule/schema"]
            if not keep_serialized:
                mols = [
                    Molecule(
                        **self._deserialize_data(mol_schema[int(i) if isinstance(i, ObjectId) else i]), validate=False
                    )
                    for i in indexes
                ]
            else:
                mols = [mol_schema[int(i) if isinstance(i, ObjectId) else i].tobytes() for i in indexes]
        return pd.Series(mols, index=indexes)

    def get_index(self, subset: Optional[List[str]] = None) -> pd.DataFrame:
        if self._index is None:
            with self._read_file() as f:
                entry_group = f["entry"]
                self._index = pd.DataFrame({"index": entry_group["entry"][()]})
                self._index["_h5idx"] = range(len(self._index))
                self._index.set_index("index", inplace=True)
        if subset is None:
            return self._index.reset_index()
        else:
            return self._index.loc[subset].reset_index()

    def get_entries(self, subset: Optional[List[str]] = None) -> pd.DataFrame:
        if self._entries is None:
            with self._read_file() as f:
                entry_group = f["entry"]
                if entry_group.attrs["model"] == "MoleculeEntry":
                    fields = ["name", "molecule_id"]
                elif entry_group.attrs["model"] == "ReactionEntry":
                    fields = ["name", "stoichiometry", "molecule", "coefficient"]
                else:
                    raise ValueError(
                        f"Unknown entry class ({entry_group.attrs['model']}) while " f"reading HDF5 entries."
                    )
                self._entries = pd.DataFrame({field: entry_group[field][()] for field in fields})
                self._entries.set_index("name", inplace=True)
        if subset is None:
            return self._entries.reset_index()
        else:
            return self._entries.loc[subset].reset_index()

    def write(self, ds: Dataset):
        # For data checksums
        dataset_kwargs = {"chunks": True, "fletcher32": True}
        ds.get_entries(force=True)
        n_records = len(ds.data.records)
        default_shape = (n_records,)

        if h5py.__version__ >= distutils.version.StrictVersion("2.10.0"):
            vlen_double_t = h5py.vlen_dtype(np.dtype("float64"))
            utf8_t = h5py.string_dtype(encoding="utf-8")
            bytes_t = h5py.vlen_dtype(np.dtype("uint8"))
            vlen_utf8_t = h5py.vlen_dtype(utf8_t)
        else:
            vlen_double_t = h5py.special_dtype(vlen=np.dtype("float64"))
            utf8_t = h5py.special_dtype(vlen=str)
            bytes_t = h5py.special_dtype(vlen=np.dtype("uint8"))
            vlen_utf8_t = h5py.special_dtype(vlen=utf8_t)

        driver_dataspec = {
            "energy": {"dtype": np.dtype("float64"), "shape": default_shape},
            "gradient": {"dtype": vlen_double_t, "shape": default_shape},
            "hessian": {"dtype": vlen_double_t, "shape": default_shape},
            "dipole": {"dtype": np.dtype("float64"), "shape": (n_records, 3)},
        }

        def _write_dataset(dataset, column, entry_dset):
            assert column.shape[1] == 1
            for i, name in enumerate(entry_dset):
                element = column.loc[name][0]
                if not h5py.check_dtype(vlen=dataset.dtype):
                    dataset[i] = element
                # Variable length datatypes require flattening of the array and special handling of missing values
                else:
                    try:
                        dataset[i] = element.ravel()
                    except AttributeError:
                        if np.isnan(element):
                            pass
                        else:
                            raise

        with self._write_file() as f:
            # Collection attributes
            for field in {
                "name",
                "collection",
                "provenance",
                "tagline",
                "tags",
                "id",
                "history_keys",
                "description",
                "metadata",
            }:
                f.attrs[field] = self._serialize_field(getattr(ds.data, field))
            if ds.client is not None:
                f.attrs["server_information"] = self._serialize_field(ds.client.server_information())
                f.attrs["server_address"] = self._serialize_field(ds.client.address)

            # Export molecules
            molecule_group = f.create_group("molecule")

            if "stoichiometry" in ds.data.history_keys:
                molecules = ds.get_molecules(stoich=list(ds.valid_stoich(force=True)), force=True)
            else:
                molecules = ds.get_molecules(force=True)
            mol_shape = (len(molecules),)
            mol_geometry = molecule_group.create_dataset(
                "geometry", shape=mol_shape, dtype=vlen_double_t, **dataset_kwargs
            )
            mol_symbols = molecule_group.create_dataset("symbols", shape=mol_shape, dtype=vlen_utf8_t, **dataset_kwargs)
            mol_schema = molecule_group.create_dataset("schema", shape=mol_shape, dtype=bytes_t, **dataset_kwargs)
            mol_charge = molecule_group.create_dataset(
                "charge", shape=mol_shape, dtype=np.dtype("float64"), **dataset_kwargs
            )
            mol_spin = molecule_group.create_dataset(
                "multiplicity", shape=mol_shape, dtype=np.dtype("int32"), **dataset_kwargs
            )
            mol_id_server_view = {}
            for i, mol_row in enumerate(molecules.to_dict("records")):
                molecule = mol_row["molecule"]
                mol_geometry[i] = molecule.geometry.ravel()
                mol_schema[i] = self._serialize_data(molecule)
                mol_symbols[i] = molecule.symbols
                mol_charge[i] = molecule.molecular_charge
                mol_spin[i] = molecule.molecular_multiplicity
                mol_id_server_view[molecule.id] = i

            # Export entries
            entry_group = f.create_group("entry")
            entry_dset = entry_group.create_dataset("entry", shape=default_shape, dtype=utf8_t, **dataset_kwargs)
            entry_dset[:] = ds.get_index(force=True)

            entries = ds.get_entries(force=True)
            if isinstance(ds.data.records[0], MoleculeEntry):
                entry_group.attrs["model"] = "MoleculeEntry"
                entries["hdf5_molecule_id"] = entries["molecule_id"].map(mol_id_server_view)
                entry_group.create_dataset("name", data=entries["name"], dtype=utf8_t, **dataset_kwargs)
                entry_group.create_dataset(
                    "molecule_id", data=entries["hdf5_molecule_id"], dtype=np.dtype("int64"), **dataset_kwargs
                )
            elif isinstance(ds.data.records[0], ReactionEntry):
                entry_group.attrs["model"] = "ReactionEntry"
                entries["hdf5_molecule_id"] = entries["molecule"].map(mol_id_server_view)
                entry_group.create_dataset("name", data=entries["name"], dtype=utf8_t, **dataset_kwargs)
                entry_group.create_dataset(
                    "stoichiometry", data=entries["stoichiometry"], dtype=utf8_t, **dataset_kwargs
                )
                entry_group.create_dataset(
                    "molecule", data=entries["hdf5_molecule_id"], dtype=np.dtype("int64"), **dataset_kwargs
                )
                entry_group.create_dataset(
                    "coefficient", data=entries["coefficient"], dtype=np.dtype("float64"), **dataset_kwargs
                )
            else:
                raise ValueError(f"Unknown entry class ({type(ds.data.records[0])}) while writing HDF5 entries.")

            # Export native data columns
            value_group = f.create_group("value")
            history = ds.list_values(native=True, force=True).reset_index().to_dict("records")
            for specification in history:
                gv_spec = specification.copy()
                name = gv_spec.pop("name")
                if "stoichiometry" in gv_spec:
                    gv_spec["stoich"] = gv_spec.pop("stoichiometry")
                dataset_name = self._normalize_hdf5_name(name)
                df = ds.get_values(name=name, force=True, native=True)
                assert df.shape[1] == 1

                driver = specification["driver"]
                dataspec = driver_dataspec[driver]
                dataset = value_group.create_dataset(dataset_name, **dataspec, **dataset_kwargs)

                for key in specification:
                    dataset.attrs[key] = self._serialize_field(specification[key])
                dataset.attrs["units"] = self._serialize_field(ds.units)

                _write_dataset(dataset, df, entry_dset)

            # Export contributed data columns
            contributed_group = f.create_group("contributed_value")
            for cv_name in ds.list_values(force=True, native=False)["name"]:
                cv_df = ds.get_values(name=cv_name, force=True, native=False)
                cv_model = ds.data.contributed_values[cv_name.lower()]

                try:
                    dataspec = driver_dataspec[cv_model.theory_level_details["driver"]]
                except (KeyError, TypeError):
                    if isinstance(cv_df[cv_name][0], float):
                        dataspec = {"dtype": np.dtype("float64"), "shape": default_shape}
                    elif isinstance(cv_df[cv_name][0], np.ndarray):
                        dataspec = {"dtype": vlen_double_t, "shape": default_shape}
                    else:
                        raise ValueError(
                            f"Unable to guess data specification for contributed value column named {cv_name}."
                        )
                    warnings.warn(
                        f"Contributed values column {cv_name} does not provide driver in theory_level_details. "
                        f"Inferred {dataspec}."
                    )

                dataset = contributed_group.create_dataset(
                    self._normalize_hdf5_name(cv_name), **dataspec, **dataset_kwargs
                )
                for field in [
                    "name",
                    "values_structure",
                    "theory_level",
                    "units",
                    "doi",
                    "external_url",
                    "citations",
                    "comments",
                    "theory_level",
                    "theory_level_details",
                ]:
                    dataset.attrs[field] = self._serialize_field(getattr(cv_model, field))

                _write_dataset(dataset, cv_df, entry_dset)

        # Clean up any caches
        self._entries = None

    def hash(self) -> str:
        """ Returns the Blake2b hash of the view """
        b2b = hashlib.blake2b()
        with open(self._path, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                b2b.update(chunk)
        return b2b.hexdigest()

    @staticmethod
    def _normalize_hdf5_name(name: str) -> str:
        """ Handles names with / in them, which is disallowed in HDF5 """
        if ":" in name:
            raise ValueError("':' not allowed in names")
        return name.replace("/", ":")

    @contextmanager
    def _read_file(self) -> Iterator["h5py.File"]:
        yield h5py.File(self._path, "r")

    @contextmanager
    def _write_file(self) -> Iterator["h5py.File"]:
        yield h5py.File(self._path, "w")

    # Methods for serializing to strings for storage in HDF5 metadata fields ("attrs")
    @staticmethod
    def _serialize_field(field: Any) -> str:
        return serialize(field, "json")

    @staticmethod
    def _deserialize_field(field: str) -> Any:
        return deserialize(field, "json")

    # Methods for serializing into HDF5 data fields
    @staticmethod
    def _serialize_data(data: Any) -> np.ndarray:
        # h5py v3 will support bytes,
        # but for now the workaround is variable-length np unit8
        return np.frombuffer(serialize(data, "msgpack-ext"), dtype="uint8")

    @staticmethod
    def _deserialize_data(data: np.ndarray) -> Any:
        return deserialize(data.tobytes(), "msgpack-ext")


class RemoteView(DatasetView):
    def __init__(self, client: "FractalClient", collection_id: int) -> None:
        """

        Parameters
        ----------
        client: FractalClient
        collection_id: int
        """
        self._client: FractalClient = client
        self._id: int = collection_id

    def get_entries(self, subset: Optional[List[str]] = None) -> pd.DataFrame:
        # TODO: consider adding a cache
        payload = {"meta": {}, "data": {"subset": subset}}

        response = self._client._automodel_request(f"collection/{self._id}/entry", "get", payload, full_return=True)
        self._check_response_meta(response.meta)
        return self._deserialize(response.data, response.meta.msgpacked_cols)

    def get_molecules(self, indexes: List[Union[ObjectId, int]]) -> pd.Series:
        payload = {"meta": {}, "data": {"indexes": indexes}}
        response = self._client._automodel_request(f"collection/{self._id}/molecule", "get", payload, full_return=True)
        self._check_response_meta(response.meta)
        df = self._deserialize(response.data, response.meta.msgpacked_cols)
        return df["molecule"].apply(lambda blob: Molecule(**blob, validate=False))

    def get_values(
        self, queries: List[Dict[str, Union[str, bool]]], subset: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Parameters
        ----------
        subset
        queries: List[Dict[str, Union[str, bool]]]
            List of queries. Fields actually used are native, name, driver
        """
        qlist = [{"name": query["name"], "driver": query["driver"], "native": query["native"]} for query in queries]
        payload = {"meta": {}, "data": {"queries": qlist, "subset": subset}}

        response = self._client._automodel_request(f"collection/{self._id}/value", "get", payload, full_return=True)
        self._check_response_meta(response.meta)
        return self._deserialize(response.data.values, response.meta.msgpacked_cols), response.data.units

    def list_values(self) -> pd.DataFrame:
        payload: Dict[str, Dict[str, Any]] = {"meta": {}, "data": {}}
        response = self._client._automodel_request(f"collection/{self._id}/list", "get", payload, full_return=True)
        self._check_response_meta(response.meta)
        return self._deserialize(response.data, response.meta.msgpacked_cols)

    def write(self, ds: Dataset) -> NoReturn:
        raise NotImplementedError()

    @staticmethod
    def _check_response_meta(meta: "CollectionSubresourceGETResponseMeta"):
        if not meta.success:
            raise RuntimeError(f"Remote view query failed with error message: {meta.error_description}")

    @staticmethod
    def _deserialize(data: bytes, msgpacked_cols: List[str]) -> pd.DataFrame:
        """
        Data are returned as feather-packed pandas DataFrames.
        Due to limitations in pyarrow, some objects are msgpacked inside the DataFrame.
        """
        import pyarrow

        df = pd.read_feather(pyarrow.BufferReader(data))
        for col in msgpacked_cols:
            df[col] = df[col].apply(lambda element: deserialize(element, "msgpack-ext"))

        if "index" in df.columns:
            df.set_index("index", inplace=True)  # pandas.to_feather does not support indexes,
            # so we have to send indexless frames over the wire, and set the index here.
        return df


class PlainTextView(DatasetView):
    def __init__(self, path: Union[str, pathlib.Path]) -> None:
        """
        Parameters
        ----------
        path: Union[str, pathlib.Path]
            File path of view
        """
        path = pathlib.Path(path)
        if len(path.suffixes) == 0:
            path = path.with_suffix(".tar.gz")
        self._path = path

    def write(self, ds: Dataset) -> None:

        with tempfile.TemporaryDirectory() as tempd:
            temppath = pathlib.Path(tempd)
            mol_path = temppath / "molecules"
            entry_path = temppath / "entries.csv"
            value_path = temppath / "values.csv"
            list_path = temppath / "value_descriptions.csv"
            readme_path = temppath / "README"

            entries = ds.get_entries(force=True)
            # calculate molecule file name
            if isinstance(ds, ReactionDataset):
                entries["molecule filename"] = [
                    normalize_filename(f"{row[1]}__{row[2]}__{row[3]}") + ".xyz" for row in entries.itertuples()
                ]
                entries.rename(columns={"molecule": "molecule_id"}, inplace=True)
            elif isinstance(ds, Dataset):
                entries["molecule filename"] = [
                    normalize_filename(f"{row[1]}__{row[2]}") + ".xyz" for row in entries.itertuples()
                ]
            else:
                raise NotImplementedError(f"Unknown dataset type: {type(ds)}.")
            entries.to_csv(entry_path)

            mol_path.mkdir()
            molecules = ds._get_molecules(
                {row[1]: row[2] for row in entries[["molecule filename", "molecule_id"]].itertuples()}, force=True
            )
            for pathname, molecule in molecules.itertuples():
                molecule.to_file(mol_path / pathname)

            ds_query_limit_state = ds._disable_query_limit
            ds._disable_query_limit = True
            ds.get_values(force=True).to_csv(value_path)
            ds._disable_query_limit = ds_query_limit_state
            df = ds.list_values(force=True).reset_index().set_index("name")
            df["units"] = ds.units
            for name in df.index:
                if name in ds._column_metadata:
                    if "units" in ds._column_metadata[name]:
                        df["units"] = ds._column_metadata[name]["units"]
            df.to_csv(list_path)

            with open(readme_path, "w") as readme_file:
                readme_file.write(self._readme(ds))

            tarpath = temppath / "archive.tar.gz"
            with tarfile.open(tarpath, "w:gz") as tarball:
                for path in [mol_path, entry_path, value_path, list_path, readme_path]:
                    tarball.add(path, arcname=path.relative_to(temppath))

            shutil.move(tarpath, self._path)

    def list_values(self) -> NoReturn:
        raise NotImplementedError()

    def get_values(self, queries: List[Dict[str, Union[str, bool]]]) -> NoReturn:
        raise NotImplementedError()

    def get_molecules(self, indexes: List[Union[ObjectId, int]]) -> NoReturn:
        raise NotImplementedError()

    def get_entries(self) -> NoReturn:
        raise NotImplementedError()

    @staticmethod
    def _readme(ds) -> str:
        # TODO: citations once we add that column
        ret = f"""Name: {ds.name}

Tagline: {ds.data.tagline}
Tags: {", ".join(ds.data.tags)}
Downloaded from: {ds.client.server_information()['name']}

Description:
{ds.data.description}

Files included:
- values.csv: Table of computed values. Rows correspond to entries (e.g. molecules, reactions). Columns correspond to methods.
- value_descriptions.csv: Table of descriptions of columns in values.csv
- entries.csv: Table of descriptions of rows in values.csv
- molecules: Folder containing XYZ-formatted geometries of molecules in the dataset. Files are named by the molecule id found in entries.csv
"""

        return ret



"""Wavefunction data table

Revision ID: 1bd2f7711e68
Revises: a2f76bb7be65
Create Date: 2019-10-17 14:03:07.364322

"""
from alembic import op
import sqlalchemy as sa
import qcfractal


# revision identifiers, used by Alembic.
revision = "1bd2f7711e68"
down_revision = "a2f76bb7be65"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "wavefunction_store",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("basis", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=False),
        sa.Column("restricted", sa.Boolean(), nullable=False),
        sa.Column("h_core_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("h_core_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("h_effective_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("h_effective_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_orbitals_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_orbitals_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_density_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_density_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_fock_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_fock_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_eigenvalues_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_eigenvalues_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_occupations_a", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("scf_occupations_b", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.Column("extras", qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.add_column("result", sa.Column("wavefunction_data_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "result_wavefunction_data_id_fkey", "result", "wavefunction_store", ["wavefunction_data_id"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("result_wavefunction_data_id_fkey", "result", type_="foreignkey")
    op.drop_column("result", "wavefunction_data_id")
    op.drop_table("wavefunction_store")
    # ### end Alembic commands ###



"""Msgpack Results Phase 2

Revision ID: 1134312ad4a3
Revises: 84c94a48e491
Create Date: 2019-08-11 17:21:43.328492

"""
from alembic import op
import sqlalchemy as sa

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers import msgpack_migrations

# revision identifiers, used by Alembic.
revision = "1134312ad4a3"
down_revision = "84c94a48e491"
branch_labels = None
depends_on = None

table_name = "result"
update_columns = {"return_result"}

nullable = {"return_result"}


def upgrade():
    msgpack_migrations.json_to_msgpack_table_altercolumns(table_name, update_columns, nullable_true=nullable)


def downgrade():
    raise ValueError("Cannot downgrade json to msgpack conversions")



"""
Import file for procedures
"""

from typing import Union, List

from .single import SingleResultTasks
from .optimization import OptimizationTasks

supported_procedures = Union[SingleResultTasks, OptimizationTasks]
__procedure_map = {"single": SingleResultTasks, "optimization": OptimizationTasks}


def check_procedure_available(procedure: str) -> List[str]:
    """
    Lists all available procedures
    """
    return procedure.lower() in __procedure_map


def get_procedure_parser(procedure_type: str, storage, logger) -> supported_procedures:
    """A factory method that returns the appropriate parser class
    for the supported procedure types (like single and optimization)
    Parameters
    ---------
    procedure_type: str, 'single' or 'optimization'
    storage: storage socket object
        such as MongoengineSocket object
    Returns
    -------
    A parser class corresponding to the procedure_type:
        'single' --> SingleResultTasks
        'optimization' --> OptimizationTasks
    """

    try:
        return __procedure_map[procedure_type.lower()](storage, logger)
    except KeyError:
        raise KeyError("Procedure type ({}) is not suported yet.".format(procedure_type))



"""
A command line interface to the qcfractal server.
"""

import os
import argparse
import shutil
import sys

import yaml

import qcfractal

from ..config import DatabaseSettings, FractalConfig, FractalServerSettings, _str2bool
from ..postgres_harness import PostgresHarness
from ..storage_sockets import storage_socket_factory
from .cli_utils import install_signal_handlers


def ensure_postgres_alive(psql):
    if not psql.is_alive():
        try:
            print("\nCould not detect a PostgreSQL from configuration options, starting a PostgreSQL server.\n")
            psql.start()
        except ValueError as e:
            print(str(e))
            sys.exit(1)


def human_sizeof_byte(num, suffix="B"):
    # SO: https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)


def standard_command_startup(name, config, check=True):
    print(f"QCFractal server {name}.\n")
    print(f"QCFractal server base folder: {config.base_folder}")

    if check:
        print("\n>>> Checking the PostgreSQL connection...")
        psql = PostgresHarness(config, quiet=False, logger=print)
        ensure_postgres_alive(psql)

    return psql


def parse_args():
    parser = argparse.ArgumentParser(description="A CLI for the QCFractalServer.")
    parser.add_argument("--version", action="version", version=f"{qcfractal.__version__}")

    subparsers = parser.add_subparsers(dest="command")

    ### Init subcommands
    init = subparsers.add_parser("init", help="Initializes a QCFractal server and database information.")
    db_init = init.add_argument_group("Database Settings")
    for field in DatabaseSettings.field_names():
        cli_name = "--db-" + field.replace("_", "-")
        db_init.add_argument(cli_name, **DatabaseSettings.help_info(field))

    server_init = init.add_argument_group("Server Settings")
    for field in FractalServerSettings.field_names():
        cli_name = "--" + field.replace("_", "-")
        server_init.add_argument(cli_name, **FractalServerSettings.help_info(field))

    init.add_argument("--overwrite-config", action="store_true", help="Overwrites the current configuration file.")
    init.add_argument(
        "--clear-database", action="store_true", help="Clear the content of the given database and initialize it."
    )
    init.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    ### Start subcommands
    start = subparsers.add_parser("start", help="Starts a QCFractal server instance.")
    start.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    # Allow some config settings to be altered via the command line
    fractal_args = start.add_argument_group("Server Settings")
    for field in ["port", "logfile", "loglevel", "cprofile"]:
        cli_name = "--" + field.replace("_", "-")
        fractal_args.add_argument(cli_name, **FractalServerSettings.help_info(field))

    fractal_args.add_argument("--server-name", **FractalServerSettings.help_info("name"))
    fractal_args.add_argument(
        "--start-periodics",
        default=True,
        type=_str2bool,
        help="Expert! Can disable periodic update (services, heartbeats) if False. Useful when running behind a proxy.",
    )

    fractal_args.add_argument(
        "--disable-ssl",
        default=False,
        type=_str2bool,
        help="Disables SSL if present, if False a SSL cert will be created for you.",
    )
    fractal_args.add_argument("--tls-cert", type=str, default=None, help="Certificate file for TLS (in PEM format)")
    fractal_args.add_argument("--tls-key", type=str, default=None, help="Private key file for TLS (in PEM format)")

    ### Upgrade subcommands
    upgrade = subparsers.add_parser("upgrade", help="Upgrade QCFractal database.")
    upgrade.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    compute_args = start.add_argument_group("Local Computation Settings")
    compute_args.add_argument(
        "--local-manager",
        const=-1,
        default=None,
        action="store",
        nargs="?",
        type=int,
        help="Creates a local pool QueueManager attached to the server.",
    )

    ### Config subcommands
    info = subparsers.add_parser("info", help="Manage users and permissions on a QCFractal server instance.")
    info.add_argument(
        "category", nargs="?", default="config", choices=["config", "alembic"], help="The config category to show."
    )
    info.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    ### User subcommands
    user = subparsers.add_parser("user", help="Configure a QCFractal server instance.")
    user.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    user_subparsers = user.add_subparsers(dest="user_command")

    user_add = user_subparsers.add_parser("add", help="Add a user to the QCFractal server.")
    user_add.add_argument("username", default=None, type=str, help="The username to add.")
    user_add.add_argument(
        "--password",
        default=None,
        type=str,
        required=False,
        help="The password for the user. If None, a default one will be created and printed.",
    )
    user_add.add_argument(
        "--permissions",
        nargs="+",
        default=None,
        type=str,
        required=True,
        help="Permissions for the user. Allowed values: read, write, queue, compute, admin.",
    )

    user_show = user_subparsers.add_parser("info", help="Show the user's current permissions.")
    user_show.add_argument("username", default=None, type=str, help="The username to show.")

    user_modify = user_subparsers.add_parser("modify", help="Change a user's password or permissions.")
    user_modify.add_argument("username", default=None, type=str, help="The username to modify.")
    user_modify_password = user_modify.add_mutually_exclusive_group()
    user_modify_password.add_argument(
        "--password", type=str, default=None, required=False, help="Change the user's password to the specified value."
    )
    user_modify_password.add_argument(
        "--reset-password",
        action="store_true",
        help="Reset the user's password. A new password will be generated and printed.",
    )
    user_modify.add_argument(
        "--permissions",
        nargs="+",
        default=None,
        type=str,
        required=False,
        help="Change the users's permissions. Allowed values: read, write, compute, queue, admin.",
    )

    user_remove = user_subparsers.add_parser("remove", help="Remove a user.")
    user_remove.add_argument("username", default=None, type=str, help="The username to remove.")

    # Backup
    backup = subparsers.add_parser("backup", help="Creates a postgres backup file of the current database.")
    backup.add_argument(
        "--filename",
        default=None,
        type=str,
        help="The filename to dump the backup to, defaults to 'database_name.bak'.",
    )
    backup.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))

    # Restore
    restore = subparsers.add_parser("restore", help="Restores the database from a backup file.")
    restore.add_argument("filename", default=None, type=str, help="The filename to restore from.")
    restore.add_argument("--base-folder", **FractalConfig.help_info("base_folder"))
    # restore.add_argument(
    #     "--force", action="store_true", help="If True, do not ask if the user wishes to delete the current database."
    # )

    ### Move args around
    args = vars(parser.parse_args())

    ret = {}
    ret["database"] = {}
    ret["fractal"] = {}
    for key, value in args.items():

        # DB bucket
        if ("db_" in key) and (key.replace("db_", "") in DatabaseSettings.field_names()):
            if value is None:
                continue
            ret["database"][key.replace("db_", "")] = value

        # Fractal bucket
        elif key in FractalServerSettings.field_names():
            if value is None:
                continue
            ret["fractal"][key] = value

        # Additional base values that should be none
        elif key in ["base_folder"]:
            if value is None:
                continue
            ret[key] = value
        else:
            ret[key] = value

    if args["command"] is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return ret


def server_init(args, config):
    # alembic stamp head

    print("Initializing QCFractal configuration.")
    # Configuration settings

    config.base_path.mkdir(parents=True, exist_ok=True)
    overwrite_config = args.get("overwrite_config", False)
    clear_database = args.get("clear_database", False)

    psql = PostgresHarness(config, quiet=False, logger=print)

    # Make sure we do not delete anything.
    if config.config_file_path.exists():
        print()
        if not overwrite_config:
            print(
                "QCFractal configuration file already exists, to overwrite use '--overwrite-config' "
                "or use the `qcfractal-server config` command line to alter settings."
            )
            sys.exit(2)
        else:
            user_required_input = f"REMOVEALLDATA {str(config.database_path)}"
            print("!WARNING! A QCFractal configuration is currently initialized")
            print(
                f"!WARNING! Overwriting will delete all current Fractal data, this includes all data in {str(config.database_path)}."
            )
            print("!WARNING! Please use `qcfractal-server config` to alter configuration settings instead.")
            print()
            print(f"!WARNING! If you are sure you wish to proceed please type '{user_required_input}' below.")

            inp = input("  > ")
            print()
            if inp == user_required_input:
                print("All data will be removed from the current QCFractal instance.")
                psql.shutdown()
                shutil.rmtree(str(config.database_path), ignore_errors=True)
            else:
                print("Input does not match 'REMOVEALLDATA', exiting.")
                sys.exit(1)

    # WARNING! Passwords do not currently work.
    # if config.database.password is None:
    #     print("  Database password is None, generating a new private key.")
    #     config.database.password = secrets.token_urlsafe(16)

    print_config = config.dict()
    print_config["database"]["password"] = "**************"
    print_config = yaml.dump(print_config, default_flow_style=False)
    print("\n>>> Settings found:\n")
    print(print_config)

    print("\n>>> Writing settings...")
    config.config_file_path.write_text(yaml.dump(config.dict(), default_flow_style=False))

    print("\n>>> Setting up PostgreSQL...\n")
    config.database_path.mkdir(exist_ok=True)
    if config.database.own:
        try:
            psql.initialize_postgres()
        except ValueError as e:
            print(str(e))
            sys.exit(1)
    else:
        print(
            "Own was set to False, QCFractal will expect a live PostgreSQL server with the above connection information."
        )

    if config.database.own or clear_database:

        print("\n>>> Initializing database schema...\n")
        try:
            psql.init_database()
        except ValueError as e:
            print(str(e))
            sys.exit(1)

    # create tables and stamp version (if not)
    print("\n>>> Finishing up...")
    print("\n>>> Success! Please run `qcfractal-server start` to boot a FractalServer!")


def server_info(args, config):

    psql = PostgresHarness(config, quiet=False, logger=print)

    if args["category"] == "config":
        print(f"Displaying QCFractal configuration:\n")
        print(yaml.dump(config.dict(), default_flow_style=False))
    elif args["category"] == "alembic":
        print(f"Displaying QCFractal Alembic CLI configuration:\n")
        print(" ".join(psql.alembic_commands()))


def server_start(args, config):
    # check if db not current, ask for upgrade

    print("Starting a QCFractal server.\n")
    print(f"QCFractal server base folder: {config.base_folder}")

    # Build an optional adapter
    if args["local_manager"]:
        ncores = args["local_manager"]
        if ncores == -1:
            ncores = None

        from concurrent.futures import ProcessPoolExecutor

        adapter = ProcessPoolExecutor(max_workers=ncores)

    else:
        adapter = None

    print("\n>>> Examining SSL Certificates...")
    # Handle SSL
    if args["disable_ssl"]:
        print("SSL disabled.")
        ssl_options = False
    else:
        ssl_certs = sum(args[x] is not None for x in ["tls_key", "tls_cert"])
        if ssl_certs == 0:
            ssl_options = True
            print("Autogenerated SSL certificates, clients must use 'verify=False' when connecting.")
        elif ssl_certs == 2:
            ssl_options = {"crt": args["tls_cert"], "key": args["tls_key"]}
            print("Reading SSL certificates.")
        else:
            raise KeyError("Both tls-cert and tls-key must be passed in.")

    # Build the server itself
    if config.fractal.logfile is None:
        logfile = None
    else:
        logfile = str(config.base_path / config.fractal.logfile)

    print("\n>>> Logging to " + logfile)
    print(">>> Loglevel: " + config.fractal.loglevel.upper())

    print("\n>>> Checking the PostgreSQL connection...")
    psql = PostgresHarness(config, quiet=False, logger=print)

    ensure_postgres_alive(psql)

    # make sure DB is created
    psql.create_database(config.database.database_name)

    print("\n>>> Initializing the QCFractal server...")
    try:
        server = qcfractal.FractalServer(
            name=args.get("server_name", None) or config.fractal.name,
            port=config.fractal.port,
            compress_response=config.fractal.compress_response,
            # Security
            security=config.fractal.security,
            allow_read=config.fractal.allow_read,
            ssl_options=ssl_options,
            # Database
            storage_uri=config.database_uri(safe=False, database=""),
            storage_project_name=config.database.database_name,
            query_limit=config.fractal.query_limit,
            # Collection views
            view_enabled=config.view.enable,
            view_path=config.view_path,
            # Log options
            logfile_prefix=logfile,
            loglevel=config.fractal.loglevel,
            log_apis=config.fractal.log_apis,
            geo_file_path=config.geo_file_path(),
            # Queue options
            service_frequency=config.fractal.service_frequency,
            heartbeat_frequency=config.fractal.heartbeat_frequency,
            max_active_services=config.fractal.max_active_services,
            queue_socket=adapter,
        )

    except Exception as e:
        print("Fatal during server startup:\n")
        print(str(e))
        print("\nFailed to start the server, shutting down.")
        sys.exit(1)

    print(f"Server: {str(server)}")

    # Register closing
    install_signal_handlers(server.loop, server.stop)

    # Blocks until keyboard interupt
    print("\n>>> Starting the QCFractal server...")
    if args["start_periodics"]:
        print("Periodics are enabled.")
    else:
        print("Periodics are disabled.")

    server.start(start_periodics=args["start_periodics"])


def server_upgrade(args, config):
    psql = standard_command_startup("upgrade", config)

    print("\n>>> Upgrading the Database...")

    try:
        psql.upgrade()
        psql.update_db_version()
    except ValueError as e:
        print(str(e))
        sys.exit(1)


def server_user(args, config):
    standard_command_startup("user function", config)

    storage = storage_socket_factory(config.database_uri(safe=False))

    try:
        if args["user_command"] == "add":
            print("\n>>> Adding new user...")
            success, pw = storage.add_user(args["username"], password=args["password"], permissions=args["permissions"])
            if success:
                print(f"\n>>> New user successfully added, password:\n{pw}")
                if config.fractal.security is None:
                    print(
                        "Warning: security is disabled. To enable security, change the configuration YAML field "
                        "fractal:security to local."
                    )
            else:
                print("\n>>> Failed to add user. Perhaps the username is already taken?")
                sys.exit(1)
        elif args["user_command"] == "info":
            print(f"\n>>> Showing permissions for user '{args['username']}'...")
            permissions = storage.get_user_permissions(args["username"])
            if permissions is None:
                print("Username not found!")
                sys.exit(1)
            else:
                print(permissions)
        elif args["user_command"] == "modify":
            print(f"\n>>> Modifying user '{args['username']}'...")
            success, message = storage.modify_user(
                args["username"], args["password"], args["reset_password"], args["permissions"]
            )
            if success:
                info = "Successfully modified user\n"
                if message is not None:
                    info += "with message: " + message
                print(info)
            else:
                print("Failed to modify user\nwith message:", message)
                sys.exit(1)
        elif args["user_command"] == "remove":
            print(f"\n>>> Removing user '{args['username']}'...")
            if storage.remove_user(args["username"]):
                print("Successfully removed user.")
            else:
                print("Failed to remove user.")
                sys.exit(1)

    except Exception as e:
        print(type(e), str(e))
        sys.exit(1)


def server_backup(args, config):
    psql = standard_command_startup("backup", config)

    print("\n>>> Starting backup, this may take several hours for large databases (100+ GB)...")

    db_size = psql.database_size()
    print(f"Current database size: {db_size['stdout']}")

    filename = args["filename"]
    if filename is None:
        filename = f"{config.database.database_name}.bak"
        print(f"No filename provided, defaulting to filename: {filename}")

    filename = os.path.realpath(filename)

    filename_temporary = filename + ".in_progress"
    psql.backup_database(filename=filename_temporary)

    shutil.move(filename_temporary, filename)

    print("Backup complete!")


def server_restore(args, config):
    psql = standard_command_startup("restore", config)

    print("!WARNING! This will erase old data. Make sure to to run 'qcfractal-server backup' before this option.")

    user_required_input = f"REMOVEALLDATA {str(config.database_path)}"
    print(f"!WARNING! If you are sure you wish to proceed please type '{user_required_input}' below.")

    inp = input("  > ")
    print()
    if inp != user_required_input:
        print(f"Input does not match '{user_required_input}', exiting.")
        sys.exit(1)

    if not os.path.isfile(args["filename"]):
        print(f"Provided filename {args['filename']} does not exist.")
        sys.exit(1)

    restore_size = os.path.getsize(args["filename"])
    human_rsize = human_sizeof_byte(restore_size)

    print("\n>>> Starting restore, this may take several hours for large databases (100+ GB)...")
    print(f"Current restore size: {human_rsize}")

    db_name = config.database.database_name
    db_backup_name = db_name + "_backup"

    print("Renaming old database for backup...")
    cmd = psql.command(f"ALTER DATABASE {db_name} RENAME TO {db_backup_name};", check=False)
    if cmd["retcode"] != 0:
        print(cmd["stderr"])
        raise ValueError("Could not rename the old database, stopping.")

    print("Restoring database...")
    try:
        psql.restore_database(args["filename"])
        psql.command(f"DROP DATABASE {db_backup_name};", check=False)
    except ValueError:
        print("Could not restore the database from file, reverting to the old database.")
        psql.command(f"DROP DATABASE {db_name};", check=False)
        psql.command(f"ALTER DATABASE {db_backup_name} RENAME TO {db_name};")
        sys.exit(1)

    print("Restore complete!")


def main(args=None):

    # Grab CLI args if not present
    if args is None:
        args = parse_args()

    command = args.pop("command")

    # More default manipulation to get supersets correct
    config_kwargs = ["base_folder", "fractal", "database"]
    config_args = {}
    for x in config_kwargs:
        if x in args:
            config_args[x] = args.pop(x)

    config = FractalConfig(**config_args)

    # If desired, enable profiling
    if config.fractal.cprofile is not None:
        print("!" * 80)
        print(f"! Enabling profiling via cProfile. Outputting data file to {config.fractal.cprofile}")
        print("!" * 80)
        import cProfile

        pr = cProfile.Profile()
        pr.enable()

    # Merge files
    if command != "init":
        if not config.base_path.exists():
            print(f"Could not find configuration file path: {config.base_path}")
            sys.exit(1)
        if not config.config_file_path.exists():
            print(f"Could not find configuration file: {config.config_file_path}")
            sys.exit(1)

        file_dict = FractalConfig(**yaml.load(config.config_file_path.read_text(), Loader=yaml.FullLoader)).dict()
        config_dict = config.dict(exclude_unset=True)

        # Only fractal options can be changed by user input parameters
        file_dict["fractal"] = {**file_dict.pop("fractal"), **config_dict.pop("fractal")}

        # base_folder is global (outside of fractal, database, etc).
        # Should be included here, just for debugging and printing purposes (as its only purpose
        # was to read the config_file above)
        file_dict["base_folder"] = config.base_folder

        config = FractalConfig(**file_dict)

    if command == "init":
        server_init(args, config)
    elif command == "info":
        server_info(args, config)
    elif command == "start":
        server_start(args, config)
    elif command == "upgrade":
        server_upgrade(args, config)
    elif command == "user":
        server_user(args, config)
    elif command == "backup":
        server_backup(args, config)
    elif command == "restore":
        server_restore(args, config)

    # Everything finished. If profiling is enabled, write out the
    # data file
    if config.fractal.cprofile is not None:
        print(f"! Writing profiling data to {config.fractal.cprofile}")
        print("! Read using the Stats class of the pstats package")
        pr.disable()
        pr.dump_stats(config.fractal.cprofile)


if __name__ == "__main__":
    main()



"""Add username to manager

Revision ID: d28e52d86633
Revises: 
Create Date: 2019-07-02 22:51:25.613742

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "d28e52d86633"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("queue_manager", sa.Column("username", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("queue_manager", "username")
    # ### end Alembic commands ###



import setuptools

import versioneer

short_description = "A distributed compute and database platform for quantum chemistry."

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except FileNotFoundError:
    long_description = short_description

if __name__ == "__main__":
    setuptools.setup(
        name="qcfractal",
        description=short_description,
        author="The QCArchive Development Team",
        author_email="qcarchive@molssi.org",
        url="https://github.com/molssi/qcfractal",
        license="BSD-3C",
        include_package_data=True,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        packages=setuptools.find_packages(),
        install_requires=[
            # Core dependencies
            "numpy",
            "msgpack >=0.6.1",
            "tornado",
            "requests",
            "pyyaml >=5.1",
            "pydantic >=1.4.0",
            # Security dependencies
            "bcrypt",
            "cryptography",
            # Storage dependencies
            "sqlalchemy >=1.3",
            "alembic",
            "psycopg2 >=2.7",
            # QCPortal dependencies
            "tqdm",
            "plotly >=4.0.0",
            "pandas",
            "h5py",
            "pyarrow >=0.13.0",
            #            'double-conversion >=3.0.0',
            # QCArchive depends
            "qcengine>=0.11.0",
            "qcelemental>=0.13.1",
        ],
        entry_points={
            "console_scripts": [
                "qcfractal-server=qcfractal.cli.qcfractal_server:main",
                "qcfractal-manager=qcfractal.cli.qcfractal_manager:main",
            ],
            "pytest11": ["qcfractal_testing=qcfractal.testing"],
        },
        extras_require={
            "api_logging": ["geoip2"],
            "docs": [
                "sphinx==1.2.3",  # autodoc was broken in 1.3.1
                "sphinxcontrib-napoleon",
                "sphinx_rtd_theme",
                "numpydoc",
            ],
            "lint": ["black", "isort"],
            "tests": ["pytest", "pytest-cov", "requests-mock"],
        },
        tests_require=["pytest", "pytest-cov"],
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
        ],
        zip_safe=True,
        long_description=long_description,
        long_description_content_type="text/markdown",
    )



"""
DQM Client base folder
"""

from . import collections, data, models, util

# Add imports here
from .client import FractalClient
from .models import Molecule

# We are running inside QCPortal repo
try:
    # The _version file exists only in the QCPortal package
    from . import _version  # lgtm [py/import-own-module]

    versions = _version.get_versions()
    __version__ = versions["version"]
    __git_revision__ = versions["full-revisionid"]
    _isportal = True

# We are running inside QCFractal
except ImportError:
    from ..extras import get_information

    __version__ = "inplace-{}".format(get_information("version"))
    __git_revision__ = get_information("git_revision")
    _isportal = False
    del get_information



"""
A model for Compute Records
"""

import abc
import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union

import numpy as np
import qcelemental as qcel
from pydantic import Field, constr, validator

from ..visualization import scatter_plot
from .common_models import DriverEnum, ObjectId, ProtoModel, QCSpecification
from .model_utils import hash_dictionary, prepare_basis, recursive_normalizer

if TYPE_CHECKING:  # pragma: no cover
    from qcelemental.models import OptimizationInput, ResultInput

    from .common_models import KeywordSet, Molecule

__all__ = ["OptimizationRecord", "ResultRecord", "OptimizationRecord", "RecordBase"]


class RecordStatusEnum(str, Enum):
    """
    The state of a record object. The states which are available are a finite set.
    """

    complete = "COMPLETE"
    incomplete = "INCOMPLETE"
    running = "RUNNING"
    error = "ERROR"


class RecordBase(ProtoModel, abc.ABC):
    """
    A BaseRecord object for Result and Procedure records. Contains all basic
    fields common to the all records.
    """

    # Classdata
    _hash_indices: Set[str]

    # Helper data
    client: Any = Field(None, description="The client object which the records are fetched from.")
    cache: Dict[str, Any] = Field(
        {},
        description="Object cache from expensive queries. It should be very rare that this needs to be set manually "
        "by the user.",
    )

    # Base identification
    id: ObjectId = Field(
        None, description="Id of the object on the database. This is assigned automatically by the database."
    )
    hash_index: Optional[str] = Field(
        None, description="Hash of this object used to detect duplication and collisions in the database."
    )
    procedure: str = Field(..., description="Name of the procedure which this Record targets.")
    program: str = Field(
        ...,
        description="The quantum chemistry program which carries out the individual quantum chemistry calculations.",
    )
    version: int = Field(..., description="The version of this record object describes.")
    protocols: Optional[Dict[str, Any]] = Field(
        None, description="Protocols that change the data stored in top level fields."
    )

    # Extra fields
    extras: Dict[str, Any] = Field({}, description="Extra information to associate with this record.")
    stdout: Optional[ObjectId] = Field(
        None,
        description="The Id of the stdout data stored in the database which was used to generate this record from the "
        "various programs which were called in the process.",
    )
    stderr: Optional[ObjectId] = Field(
        None,
        description="The Id of the stderr data stored in the database which was used to generate this record from the "
        "various programs which were called in the process.",
    )
    error: Optional[ObjectId] = Field(
        None,
        description="The Id of the error data stored in the database in the event that an error was generated in the "
        "process of carrying out the process this record targets. If no errors were raised, this field "
        "will be empty.",
    )

    # Compute status
    manager_name: Optional[str] = Field(None, description="Name of the Queue Manager which generated this record.")
    status: RecordStatusEnum = Field(RecordStatusEnum.incomplete, description=str(RecordStatusEnum.__doc__))
    modified_on: datetime.datetime = Field(None, description="Last time the data this record points to was modified.")
    created_on: datetime.datetime = Field(None, description="Time the data this record points to was first created.")

    # Carry-ons
    provenance: Optional[qcel.models.Provenance] = Field(
        None,
        description="Provenance information tied to the creation of this record. This includes things such as every "
        "program which was involved in generating the data for this record.",
    )

    class Config(ProtoModel.Config):
        build_hash_index = True

    @validator("program")
    def check_program(cls, v):
        return v.lower()

    def __init__(self, **data):

        # Set datetime defaults if not automatically available
        data.setdefault("modified_on", datetime.datetime.utcnow())
        data.setdefault("created_on", datetime.datetime.utcnow())

        super().__init__(**data)

        # Set hash index if not present
        if self.Config.build_hash_index and (self.hash_index is None):
            self.__dict__["hash_index"] = self.get_hash_index()

    def __repr_args__(self):

        return [("id", f"{self.id}"), ("status", f"{self.status}")]

    ### Serialization helpers

    @classmethod
    def get_hash_fields(cls) -> Set[str]:
        """Provides a description of the fields to be used in the hash
        that uniquely defines this object.

        Returns
        -------
        Set[str]
            A list of all fields that are used in the hash.

        """
        return cls._hash_indices | {"procedure", "program"}

    def get_hash_index(self) -> str:
        """Builds (or rebuilds) the hash of this
        object using the internally known hash fields.

        Returns
        -------
        str
            The objects unique hash index.
        """
        data = self.dict(include=self.get_hash_fields(), encoding="json")

        return hash_dictionary(data)

    def dict(self, *args, **kwargs):
        kwargs["exclude"] = (kwargs.pop("exclude", None) or set()) | {"client", "cache"}
        # kwargs["skip_defaults"] = True
        return super().dict(*args, **kwargs)

    ### Checkers

    def check_client(self, noraise: bool = False) -> bool:
        """Checks whether this object owns a FractalClient or not.
        This is often done so that objects pulled from a server using
        a FractalClient still posses a connection to the server so that
        additional data related to this object can be queried.

        Raises
        ------
        ValueError
            If this object does not contain own a client.

        Parameters
        ----------
        noraise : bool, optional
            Does not raise an error if this is True and instead returns
            a boolean depending if a client exists or not.

        Returns
        -------
        bool
            If True, the object owns a connection to a server. False otherwise.
        """
        if self.client is None:
            if noraise:
                return False

            raise ValueError("Requested method requires a client, but client was '{}'.".format(self.client))

        return True

    ### KVStore Getters

    def _kvstore_getter(self, field_name):
        """
        Internal KVStore getting object
        """
        self.check_client()

        oid = self.__dict__[field_name]
        if oid is None:
            return None

        if field_name not in self.cache:
            # Decompress here, rather than later
            # that way, it is decompressed in the cache
            kv = self.client.query_kvstore([oid])[oid]

            if field_name == "error":
                self.cache[field_name] = kv.get_json()
            else:
                self.cache[field_name] = kv.get_string()

        return self.cache[field_name]

    def get_stdout(self) -> Optional[str]:
        """Pulls the stdout from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[str]
            The requested stdout, none if no stdout present.
        """
        return self._kvstore_getter("stdout")

    def get_stderr(self) -> Optional[str]:
        """Pulls the stderr from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[str]
            The requested stderr, none if no stderr present.
        """

        return self._kvstore_getter("stderr")

    def get_error(self) -> Optional[qcel.models.ComputeError]:
        """Pulls the stderr from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[qcel.models.ComputeError]
            The requested compute error, none if no error present.
        """
        value = self._kvstore_getter("error")
        if value:
            return qcel.models.ComputeError(**value)
        else:
            return value


class ResultRecord(RecordBase):

    # Classdata
    _hash_indices = {"driver", "method", "basis", "molecule", "keywords", "program"}

    # Version data
    version: int = Field(1, description="Version of the ResultRecord Model which this data was created with.")
    procedure: constr(strip_whitespace=True, regex="single") = Field(
        "single", description='Procedure is fixed as "single" because this is single quantum chemistry result.'
    )

    # Input data
    driver: DriverEnum = Field(..., description=str(DriverEnum.__doc__))
    method: str = Field(..., description="The quantum chemistry method the driver runs with.")
    molecule: ObjectId = Field(
        ..., description="The Id of the molecule in the Database which the result is computed on."
    )
    basis: Optional[str] = Field(
        None,
        description="The quantum chemistry basis set to evaluate (e.g., 6-31g, cc-pVDZ, ...). Can be ``None`` for "
        "methods without basis sets.",
    )
    keywords: Optional[ObjectId] = Field(
        None,
        description="The Id of the :class:`KeywordSet` which was passed into the quantum chemistry program that "
        "performed this calculation.",
    )
    protocols: Optional[qcel.models.results.ResultProtocols] = Field(
        qcel.models.results.ResultProtocols(), description=""
    )

    # Output data
    return_result: Union[float, qcel.models.types.Array[float], Dict[str, Any]] = Field(
        None, description="The primary result of the calculation, output is a function of the specified ``driver``."
    )
    properties: qcel.models.ResultProperties = Field(
        None, description="Additional data and results computed as part of the ``return_result``."
    )
    wavefunction: Optional[Dict[str, Any]] = Field(None, description="Wavefunction data generated by the Result.")
    wavefunction_data_id: Optional[ObjectId] = Field(None, description="The id of the wavefunction")

    class Config(RecordBase.Config):
        """A hash index is not used for ResultRecords as they can be
        uniquely determined with queryable keys.
        """

        build_hash_index = False

    @validator("method")
    def check_method(cls, v):
        """Methods should have a lower string to match the database."""
        return v.lower()

    @validator("basis")
    def check_basis(cls, v):
        return prepare_basis(v)

    def get_wavefunction(self, key: Union[str, List[str]]) -> Any:
        """
        Pulls down the Wavefunction data associated with the computation.
        """

        if self.wavefunction is None:
            raise AttributeError("This Record was not computed with Wavefunction data.")

        single_return = False
        if isinstance(key, str):
            key = [key]
            single_return = True

        keys = [x.lower() for x in key]

        self.cache.setdefault("wavefunction", {})

        mapped_keys = {self.wavefunction["return_map"].get(x, x) for x in keys}
        missing = mapped_keys - self.cache["wavefunction"].keys()

        unknown = missing - set(self.wavefunction["available"] + ["basis", "restricted"])
        if unknown:
            raise KeyError(
                f"Wavefunction Key(s) `{unknown}` not understood, available keys are: {self.wavefunction['available']}"
            )

        if missing:

            # Translate a return value
            proj = [self.wavefunction["return_map"].get(x, x) for x in missing]

            self.cache["wavefunction"].update(
                self.client.custom_query(
                    "wavefunctionstore", None, {"id": self.wavefunction_data_id}, meta={"include": proj}
                )
            )

            if "basis" in missing:
                self.cache["wavefunction"]["basis"] = qcel.models.BasisSet(**self.cache["wavefunction"]["basis"])

        # Remap once more
        ret = {}
        for k in keys:
            mkey = self.wavefunction["return_map"].get(k, k)
            ret[k] = self.cache["wavefunction"][mkey]

        if single_return:
            return ret[keys[0]]
        else:
            return ret

    def get_molecule(self) -> "Molecule":
        """
        Pulls the Result's Molecule from the connected database.

        Returns
        -------
        Molecule
            The requested Molecule
        """
        self.check_client()

        if self.molecule is None:
            return None

        if "molecule" not in self.cache:
            self.cache["molecule"] = self.client.query_molecules(id=self.molecule)[0]

        return self.cache["molecule"]


class OptimizationRecord(RecordBase):
    """
    A OptimizationRecord for all optimization procedure data.
    """

    # Class data
    _hash_indices = {"initial_molecule", "keywords", "qc_spec"}

    # Version data
    version: int = Field(1, description="Version of the OptimizationRecord Model which this data was created with.")
    procedure: constr(strip_whitespace=True, regex="optimization") = Field(
        "optimization", description='A fixed string indication this is a record for an "Optimization".'
    )
    schema_version: int = Field(1, description="The version number of QCSchema under which this record conforms to.")

    # Input data
    initial_molecule: ObjectId = Field(
        ..., description="The Id of the molecule which was passed in as the reference for this Optimization."
    )
    qc_spec: QCSpecification = Field(
        ..., description="The specification of the quantum chemistry calculation to run at each point."
    )
    keywords: Dict[str, Any] = Field(
        {},
        description="The keyword options which were passed into the Optimization program. "
        "Note: These are a dictionary and not a :class:`KeywordSet` object.",
    )
    protocols: Optional[qcel.models.procedures.OptimizationProtocols] = Field(
        qcel.models.procedures.OptimizationProtocols(), description=""
    )

    # Automatting issue currently
    # description=str(qcel.models.procedures.OptimizationProtocols.__doc__))

    # Results
    energies: List[float] = Field(None, description="The ordered list of energies at each step of the Optimization.")
    final_molecule: ObjectId = Field(
        None, description="The ``ObjectId`` of the final, optimized Molecule the Optimization procedure converged to."
    )
    trajectory: List[ObjectId] = Field(
        None,
        description="The list of Molecule Id's the Optimization procedure generated at each step of the optimization."
        "``initial_molecule`` will be the first index, and ``final_molecule`` will be the last index.",
    )

    class Config(RecordBase.Config):
        pass

    @validator("keywords")
    def check_keywords(cls, v):
        if v is not None:
            v = recursive_normalizer(v)
        return v

    ## Standard function

    def get_final_energy(self) -> float:
        """The final energy of the geometry optimization.

        Returns
        -------
        float
            The optimization molecular energy.
        """
        return self.energies[-1]

    def get_trajectory(self) -> List[ResultRecord]:
        """Returns the Result records for each gradient evaluation in the trajectory.

        Returns
        -------
        List['ResultRecord']
            A ordered list of Result record gradient computations.

        """

        if "trajectory" not in self.cache:
            result = {x.id: x for x in self.client.query_results(id=self.trajectory)}

            self.cache["trajectory"] = [result[x] for x in self.trajectory]

        return self.cache["trajectory"]

    def get_molecular_trajectory(self) -> List["Molecule"]:
        """Returns the Molecule at each gradient evaluation in the trajectory.

        Returns
        -------
        List['Molecule']
            A ordered list of Molecules in the trajectory.

        """

        if "molecular_trajectory" not in self.cache:
            mol_ids = [x.molecule for x in self.get_trajectory()]

            mols = {x.id: x for x in self.client.query_molecules(id=mol_ids)}
            self.cache["molecular_trajectory"] = [mols[x] for x in mol_ids]

        return self.cache["molecular_trajectory"]

    def get_initial_molecule(self) -> "Molecule":
        """Returns the initial molecule

        Returns
        -------
        Molecule
            The initial molecule
        """

        ret = self.client.query_molecules(id=[self.initial_molecule])
        return ret[0]

    def get_final_molecule(self) -> "Molecule":
        """Returns the optimized molecule

        Returns
        -------
        Molecule
            The optimized molecule
        """

        ret = self.client.query_molecules(id=[self.final_molecule])
        return ret[0]

    ## Show functions

    def show_history(
        self, units: str = "kcal/mol", digits: int = 3, relative: bool = True, return_figure: Optional[bool] = None
    ) -> "plotly.Figure":
        """Plots the energy of the trajectory the optimization took.

        Parameters
        ----------
        units : str, optional
            Units to display the trajectory in.
        digits : int, optional
            The number of valid digits to show.
        relative : bool, optional
            If True, all energies are shifted by the lowest energy in the trajectory. Otherwise provides raw energies.
        return_figure : Optional[bool], optional
            If True, return the raw plotly figure. If False, returns a hosted iPlot. If None, return a iPlot display in
            Jupyter notebook and a raw plotly figure in all other circumstances.

        Returns
        -------
        plotly.Figure
            The requested figure.
        """
        cf = qcel.constants.conversion_factor("hartree", units)

        energies = np.array(self.energies)
        if relative:
            energies = energies - np.min(energies)

        trace = {"mode": "lines+markers", "x": list(range(1, len(energies) + 1)), "y": np.around(energies * cf, digits)}

        if relative:
            ylabel = f"Relative Energy [{units}]"
        else:
            ylabel = f"Absolute Energy [{units}]"

        custom_layout = {
            "title": "Geometry Optimization",
            "yaxis": {"title": ylabel, "zeroline": True},
            "xaxis": {
                "title": "Optimization Step",
                # "zeroline": False,
                "range": [min(trace["x"]), max(trace["x"])],
            },
        }

        return scatter_plot([trace], custom_layout=custom_layout, return_figure=return_figure)



"""Remove task_id from base_results table

Revision ID: dd35678aae90
Revises: 30fd8253d87f
Create Date: 2020-10-26 16:35:40.136241

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "dd35678aae90"
down_revision = "30fd8253d87f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("base_result", "task_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("base_result", sa.Column("task_id", sa.VARCHAR(), autoincrement=False, nullable=True))
    # ### end Alembic commands ###



"""
Mongo QCDB Abstract basic Collection class

Helper
"""

import abc
import copy
import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union

import pandas as pd

from ..models import ProtoModel

if TYPE_CHECKING:  # pragma: no cover
    from .. import FractalClient
    from ..models import ObjectId


class Collection(abc.ABC):
    def __init__(self, name: str, client: Optional["FractalClient"] = None, **kwargs: Any):
        """
        Initializer for the Collection objects. If no Portal is supplied or the Collection name
        is not present on the server that the Portal is connected to a blank Collection will be
        created.

        Parameters
        ----------
        name : str
            The name of the Collection object as ID'ed on the storage backend.
        client : FractalClient, optional
            A FractalClient connected to a server
        **kwargs : Dict[str, Any]
            Additional keywords which are passed to the Collection and the initial data constructor
            It is up to the individual implementations of the Collection to do things with that data
        """

        self.client = client
        if (self.client is not None) and not (self.client.__class__.__name__ == "FractalClient"):
            raise TypeError("Expected FractalClient as `client` kwarg, found {}.".format(type(self.client)))

        if "collection" not in kwargs:
            kwargs["collection"] = self.__class__.__name__.lower()

        kwargs["name"] = name

        # Create the data model
        self.data = self.DataModel(**kwargs)

    class DataModel(ProtoModel):
        """
        Internal Data structure base model typed by PyDantic

        This structure validates input, allows server-side validation and data security,
        and will create the information to pass back and forth between server and client

        Subclasses of Collection can extend this class internally to change the set of
        additional data defined by the Collection
        """

        id: str = "local"
        name: str

        collection: str
        provenance: Dict[str, str] = {}

        tags: List[str] = []
        tagline: Optional[str] = None
        description: Optional[str] = None

        group: str = "default"
        visibility: bool = True

        view_url_hdf5: Optional[str] = None
        view_url_plaintext: Optional[str] = None
        view_metadata: Optional[Dict[str, str]] = None
        view_available: bool = False

        metadata: Dict[str, Any] = {}

    def __str__(self) -> str:
        """
        A simple string representation of the Collection.

        Returns
        -------
        ret : str
            A representation of the Collection.

        Examples
        --------

        >>> repr(obj)
        Collection(name=`S22`, id='5b7f1fd57b87872d2c5d0a6d', client=`localhost:8888`)
        """

        client = None
        if self.client:
            client = self.client.address

        class_name = self.__class__.__name__
        ret = "{}(".format(class_name)
        ret += "name=`{}`, ".format(self.data.name)
        ret += "id='{}', ".format(self.data.id)
        ret += "client='{}') ".format(client)

        return ret

    def __repr__(self) -> str:
        return f"<{self}>"

    def _check_client(self):
        if self.client is None:
            raise AttributeError("This method requires a FractalClient and no client was set")

    @property
    def name(self) -> str:
        return self.data.name

    @classmethod
    def from_server(cls, client: "FractalClient", name: str) -> "Collection":
        """Creates a new class from a server

        Parameters
        ----------
        client : FractalClient
            A FractalClient connected to a server
        name : str
            The name of the collection to pull from.

        Returns
        -------
        Collection
            A constructed collection.

        """

        if not (client.__class__.__name__ == "FractalClient"):
            raise TypeError("Expected a FractalClient as first argument, found {}.".format(type(client)))

        class_name = cls.__name__.lower()
        tmp_data = client.get_collection(class_name, name, full_return=True)
        if tmp_data.meta.n_found == 0:
            raise KeyError("Warning! `{}: {}` not found.".format(class_name, name))

        return cls.from_json(tmp_data.data[0], client=client)

    @classmethod
    def from_json(cls, data: Dict[str, Any], client: "FractalClient" = None) -> "Collection":
        """Creates a new class from a JSON blob

        Parameters
        ----------
        data : Dict[str, Any]
            The JSON blob to create a new class from.
        client : FractalClient, optional
            A FractalClient connected to a server

        Returns
        -------
        Collection
            A constructed collection.

        """
        # Check we are building the correct object
        class_name = cls.__name__.lower()
        if "collection" not in data:
            raise KeyError("Attempted to create Collection from JSON, but no `collection` field found.")

        if data["collection"] != class_name:
            raise KeyError(
                "Attempted to create Collection from JSON with class {}, but found collection type of {}.".format(
                    class_name, data["collection"]
                )
            )

        name = data.pop("name")
        # Allow PyDantic to handle type validation
        ret = cls(name, client=client, **data)
        return ret

    def to_json(self, filename: Optional[str] = None):
        """
        If a filename is provided, dumps the file to disk. Otherwise returns a copy of the current data.

        Parameters
        ----------
        filename : str, Optional, Default: None
            The filename to drop the data to.

        Returns
        -------
        ret : dict
            A JSON representation of the Collection
        """
        data = self.data.dict()
        if filename is not None:
            with open(filename, "w") as open_file:
                json.dump(data, open_file)
        else:
            return copy.deepcopy(data)

    @abc.abstractmethod
    def _pre_save_prep(self, client: "FractalClient"):
        """
        Additional actions to take before saving, done as the last step before data is written.

        This does not return anything but can prep the `self.data` field before storing it.

        Has access to the `client` in case its needed to do pre-conditioning.

        Parameters
        ----------
        client : FractalClient
            A FractalClient connected to a server used for storage access
        """

    # Setters
    def save(self, client: Optional["FractalClient"] = None) -> "ObjectId":
        """Uploads the overall structure of the Collection (indices, options, new molecules, etc)
        to the server.

        Parameters
        ----------
        client : FractalClient, optional
            A FractalClient connected to a server to upload to

        Returns
        -------
        ObjectId
            The ObjectId of the saved collection.

        """
        class_name = self.__class__.__name__.lower()
        if self.data.name == "":
            raise AttributeError("Collection:save: {} must have a name!".format(class_name))

        if client is None:
            self._check_client()
            client = self.client

        self._pre_save_prep(client)

        # Add the database
        if self.data.id == self.data.__fields__["id"].default:
            response = client.add_collection(self.data.dict(), overwrite=False, full_return=True)
            if response.meta.success is False:
                raise KeyError(f"Error adding collection: \n{response.meta.error_description}")
            self.data.__dict__["id"] = response.data
        else:
            response = client.add_collection(self.data.dict(), overwrite=True, full_return=True)
            if response.meta.success is False:
                raise KeyError(f"Error updating collection: \n{response.meta.error_description}")

        return self.data.id

    ### General helpers

    @staticmethod
    def _add_molecules_by_dict(client, molecules):

        flat_map_keys = []
        flat_map_mols = []
        for k, v in molecules.items():
            flat_map_keys.append(k)
            flat_map_mols.append(v)

        CHUNK_SIZE = client.query_limit
        mol_ret = []
        for i in range(0, len(flat_map_mols), CHUNK_SIZE):
            mol_ret.extend(client.add_molecules(flat_map_mols[i : i + CHUNK_SIZE]))

        return {k: v for k, v in zip(flat_map_keys, mol_ret)}


class BaseProcedureDataset(Collection):
    def __init__(self, name: str, client: "FractalClient" = None, **kwargs):
        if client is None:
            raise KeyError("{self.__class__.__name__} must initialize with a client.")

        super().__init__(name, client=client, **kwargs)

        self.df = pd.DataFrame(index=self._get_index())

    class DataModel(Collection.DataModel):

        records: Dict[str, Any] = {}
        history: Set[str] = set()
        specs: Dict[str, Any] = {}

        class Config(Collection.DataModel.Config):
            pass

    @abc.abstractmethod
    def _internal_compute_add(self, spec: Any, entry: Any, tag: str, priority: str) -> "ObjectId":
        pass

    def _pre_save_prep(self, client: "FractalClient") -> None:
        pass

    def _get_index(self):

        return [x.name for x in self.data.records.values()]

    def _add_specification(self, name: str, spec: Any, overwrite=False) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the specification
        spec : Any
            The specification object
        overwrite : bool, optional
            Overwrite existing specification names

        """

        lname = name.lower()
        if (lname in self.data.specs) and (not overwrite):
            raise KeyError(f"{self.__class__.__name__} '{name}' already present, use `overwrite=True` to replace.")

        self.data.specs[lname] = spec
        self.save()

    def _get_procedure_ids(self, spec: str, sieve: Optional[List[str]] = None) -> Dict[str, "ObjectId"]:
        """Aquires the

        Parameters
        ----------
        spec : str
            The specification to get the map of
        sieve : Optional[List[str]], optional
            A
            Description

        Returns
        -------
        Dict[str, ObjectId]
            A dictionary of identifier to id mappings.

        """

        spec = self.get_specification(spec)

        mapper = {}
        for rec in self.data.records.values():
            if sieve and rec.name not in sieve:
                continue

            try:
                td_id = rec.object_map[spec.name]
                mapper[rec.name] = td_id
            except KeyError:
                pass

        return mapper

    def get_specification(self, name: str) -> Any:
        """
        Parameters
        ----------
        name : str
            The name of the specification

        Returns
        -------
        Specification
            The requested specification.

        """
        try:
            return self.data.specs[name.lower()].copy()
        except KeyError:
            raise KeyError(f"Specification '{name}' not found.")

    def list_specifications(self, description=True) -> Union[List[str], pd.DataFrame]:
        """Lists all available specifications

        Parameters
        ----------
        description : bool, optional
            If True returns a DataFrame with
            Description

        Returns
        -------
        Union[List[str], 'DataFrame']
            A list of known specification names.

        """
        if description:
            data = [(x.name, x.description) for x in self.data.specs.values()]
            return pd.DataFrame(data, columns=["Name", "Description"]).set_index("Name")
        else:
            return [x.name for x in self.data.specs.values()]

    def _check_entry_exists(self, name):
        """
        Checks if an entry exists or not.
        """

        if name.lower() in self.data.records:
            raise KeyError(f"Record {name} already in the dataset.")

    def _add_entry(self, name, record, save):
        """
        Adds an entry to the records
        """

        self._check_entry_exists(name)
        self.data.records[name.lower()] = record
        if save:
            self.save()

    def get_entry(self, name: str) -> Any:
        """Obtains a record from the Dataset

        Parameters
        ----------
        name : str
            The record name to pull from.

        Returns
        -------
        Record
            The requested record
        """
        try:
            return self.data.records[name.lower()]
        except KeyError:
            raise KeyError(f"Could not find entry name '{name}' in the dataset.")

    def get_record(self, name: str, specification: str) -> Any:
        """Pulls an individual computational record of the requested name and column.

        Parameters
        ----------
        name : str
            The index name to pull the record of.
        specification : str
            The name of specification to pull the record of.

        Returns
        -------
        Any
            The requested Record

        """
        spec = self.get_specification(specification)
        rec_id = self.get_entry(name).object_map.get(spec.name, None)

        if rec_id is None:
            raise KeyError(f"Could not find a record for ({name}: {specification}).")

        return self.client.query_procedures(id=rec_id)[0]

    def compute(
        self, specification: str, subset: Set[str] = None, tag: Optional[str] = None, priority: Optional[str] = None
    ) -> int:
        """Computes a specification for all entries in the dataset.

        Parameters
        ----------
        specification : str
            The specification name.
        subset : Set[str], optional
            Computes only a subset of the dataset.
        tag : Optional[str], optional
            The queue tag to use when submitting compute requests.
        priority : Optional[str], optional
            The priority of the jobs low, medium, or high.

        Returns
        -------
        int
            The number of submitted computations
        """

        specification = specification.lower()
        spec = self.get_specification(specification)
        if subset:
            subset = set(subset)

        submitted = 0
        for entry in self.data.records.values():
            if (subset is not None) and (entry.name not in subset):
                continue

            if spec.name in entry.object_map:
                continue

            entry.object_map[spec.name] = self._internal_compute_add(spec, entry, tag, priority)
            submitted += 1

        self.data.history.add(specification)

        # Nothing to save
        if submitted:
            self.save()

        return submitted

    def query(self, specification: str, force: bool = False) -> pd.Series:
        """Queries a given specification from the server

        Parameters
        ----------
        specification : str
            The specification name to query
        force : bool, optional
            Force a fresh query if the specification already exists.

        Returns
        -------
        pd.Series
            Records collected from the server
        """
        # Try to get the specification, will throw if not found.
        spec = self.get_specification(specification)

        if not force and (spec.name in self.df):
            return spec.name

        mapper = self._get_procedure_ids(spec.name)
        query_ids = list(mapper.values())

        # Chunk up the queries
        procedures: List[Dict[str, Any]] = []
        for i in range(0, len(query_ids), self.client.query_limit):
            chunk_ids = query_ids[i : i + self.client.query_limit]
            procedures.extend(self.client.query_procedures(id=chunk_ids))

        proc_lookup = {x.id: x for x in procedures}

        data = []
        for name, oid in mapper.items():
            try:
                data.append([name, proc_lookup[oid]])
            except KeyError:
                data.append([name, None])

        df = pd.DataFrame(data, columns=["index", spec.name])
        df.set_index("index", inplace=True)

        self.df[spec.name] = df[spec.name]

        return df[spec.name]

    def status(
        self,
        specs: Union[str, List[str]] = None,
        collapse: bool = True,
        status: Optional[str] = None,
        detail: bool = False,
    ) -> pd.DataFrame:
        """Returns the status of all current specifications.

        Parameters
        ----------
        collapse : bool, optional
            Collapse the status into summaries per specification or not.
        status : Optional[str], optional
            If not None, only returns results that match the provided status.
        detail : bool, optional
            Shows a detailed description of the current status of incomplete jobs.

        Returns
        -------
        DataFrame
            A DataFrame of all known statuses

        """

        # Simple no detail case
        if detail is False:
            # detail = False can handle multiple specifications
            # If specs is None, then use all (via list_specifications)
            if isinstance(specs, str):
                specs = [specs]
            elif specs is None:
                specs = self.list_specifications(description=False)

            # Query all of the specs and make sure they are valid
            # Specs may not be loaded to self.df yet. This can be accomplished
            #     with self.query, which stores the info in self.df
            for spec in specs:
                self.query(spec)

            def get_status(item):
                try:
                    return item.status.value
                except AttributeError:
                    return None

            # apply status by column then by row
            df = self.df[specs].apply(lambda col: col.apply(get_status))

            if status:
                df = df[(df == status.upper()).all(axis=1)]

            if collapse:
                return df.apply(lambda x: x.value_counts())
            else:
                return df

        if status not in [None, "INCOMPLETE"]:
            raise KeyError("Detailed status is only available for incomplete procedures.")

        # Can only do detailed status for a single spec
        # If specs is a string, ok. If it is a list, then it should have length = 1
        if not (isinstance(specs, str) or len(specs) == 1):
            raise KeyError("Detailed status is only available for a single specification at a time.")

        # If specs is a list (of length = 1, checked above), then make it a string
        # (_get_procedure_ids expects a string)
        if not isinstance(specs, str):
            specs = specs[0]

        mapper = self._get_procedure_ids(specs)
        reverse_map = {v: k for k, v in mapper.items()}
        procedures = self.client.query_procedures(id=list(mapper.values()))

        data = []

        for proc in procedures:
            if proc.status == "COMPLETE":
                continue

            try:
                blob = proc.detailed_status()
            except:
                raise AttributeError("Detailed statuses are not available for this dataset type.")

            blob["Name"] = reverse_map[proc.id]
            data.append(blob)

        df = pd.DataFrame(data)
        df.rename(columns={x: x.replace("_", " ").title() for x in df.columns}, inplace=True)
        if df.shape[0]:
            df = df.set_index("Name")

        return df
